""" test the particle mass calculation
"""

import numpy as np
import pytest
from particula import u
from particula.util.particle_surface import area


def test_particle_mass():
    """ Testing the particle area calculation:

            * check defaults
            * check raise errors
            * check output units/mag
            * ensures radius is mandatory
    """

    a_radius = 1 * u.m

    b_radius = a_radius.m

    a_area = area(radius=a_radius)
    b_area = area(radius=b_radius)

    assert a_area.magnitude == 4 * np.pi
    assert a_area.units == u.m**2
    assert b_area == a_area

    assert area(radius=[1e-9, 1e-8]).m.shape == (2,)
    assert area(radius=1-9, area_factor=[1, 0.9]).m.shape == (2,)
    assert area(radius=[1e-9, 1e-8], area_factor=[1, 0.9]).m.shape == (2,)

    with pytest.raises(ValueError):
        area(radius=1*u.kg)

    with pytest.raises(ValueError):
        area(radius=1*u.m, area_factor=1*u.kg)

    with pytest.raises(ValueError):
        area()



"""Tests if the binary activity is working, does not
verify values, just that it can be called."""

import numpy as np

from particula.activity import binary_activity


def test_convert_to_oh_eqivalent():
    """test for convert_to_oh_equivalent function."""

    molar_mass_ratio = 18.016 / 250
    oxygen2carbon = 0.3

    new_oxygen2carbon, new_molar_mass_ratio = \
        binary_activity.convert_to_oh_equivalent(
            oxygen2carbon=oxygen2carbon,
            molar_mass_ratio=molar_mass_ratio,
            functional_group=None
        )
    assert new_oxygen2carbon == oxygen2carbon
    assert new_molar_mass_ratio == molar_mass_ratio


def test_bat_blending_weights():
    """test for bat_blending_weights function."""
    molar_mass_ratio = 18.016 / 250
    oxygen2carbon = 0.3

    weights = binary_activity.bat_blending_weights(
        molar_mass_ratio=molar_mass_ratio,
        oxygen2carbon=oxygen2carbon
    )
    assert np.all(weights >= 0)
    assert np.all(weights <= 1)


def test_biphasic_water_activity_point():
    """test for biphasic_water_activity_point function."""
    oxygen2carbon = 0.3
    hydrogen2carbon = 0.3
    molar_mass_ratio = 18/250

    # single entry
    activity_point = binary_activity.biphasic_water_activity_point(
        oxygen2carbon=oxygen2carbon,
        hydrogen2carbon=hydrogen2carbon,
        molar_mass_ratio=molar_mass_ratio,
        functional_group=None,
    )
    assert np.all(activity_point >= 0)

    # array entry
    oxygen2carbon_array = np.linspace(0, 0.6, 10)
    hydrogen2carbon_array = np.linspace(0, 0.6, 10)
    molar_mass_ratio_array = 18/np.linspace(200, 500, 10)
    activity_point = binary_activity.biphasic_water_activity_point(
        oxygen2carbon=oxygen2carbon_array,
        hydrogen2carbon=hydrogen2carbon_array,
        molar_mass_ratio=molar_mass_ratio_array,
        functional_group=None,
    )
    assert np.all(activity_point >= 0)


def test_gibbs_of_mixing():
    """test for gibbs_of_mixing function."""
    molar_mass_ratio = 18.016 / 250
    organic_mole_fraction = np.linspace(0.1, 1, 10)
    oxygen2carbon = 0.3
    density = 2.5

    gibbs_mix, dervative_gibbs = binary_activity.gibbs_of_mixing(
        molar_mass_ratio=molar_mass_ratio,
        organic_mole_fraction=organic_mole_fraction,
        oxygen2carbon=oxygen2carbon,
        density=density,
        fit_dict=binary_activity.FIT_LOW,
    )
    assert np.all(gibbs_mix >= 0)
    assert np.all(dervative_gibbs ** 2 >= 0)

    # repeat for mid fit
    gibbs_mix, dervative_gibbs = binary_activity.gibbs_of_mixing(
        molar_mass_ratio=molar_mass_ratio,
        organic_mole_fraction=organic_mole_fraction,
        oxygen2carbon=oxygen2carbon,
        density=density,
        fit_dict=binary_activity.FIT_MID,
    )
    assert np.all(gibbs_mix >= 0)
    assert np.all(dervative_gibbs ** 2 >= 0)

    # repeat for high fit
    gibbs_mix, dervative_gibbs = binary_activity.gibbs_of_mixing(
        molar_mass_ratio=molar_mass_ratio,
        organic_mole_fraction=organic_mole_fraction,
        oxygen2carbon=oxygen2carbon,
        density=density,
        fit_dict=binary_activity.FIT_HIGH,
    )
    assert np.all(gibbs_mix >= 0)
    assert np.all(dervative_gibbs ** 2 >= 0)


def test_activity_coefficents():
    """test for activity_coefficents function."""
    molar_mass_ratio = 18.016 / 250
    organic_mole_fraction = np.linspace(0.1, 1, 10)
    oxygen2carbon = 0.3
    density = 2.5

    activity_coefficients = binary_activity.activity_coefficients(
        molar_mass_ratio=molar_mass_ratio,
        organic_mole_fraction=organic_mole_fraction,
        oxygen2carbon=oxygen2carbon,
        density=density,
        functional_group=None,
    )
    assert np.all(activity_coefficients[0] >= 0)
    assert np.all(activity_coefficients[1] >= 0)
    assert np.all(activity_coefficients[2] >= 0)
    assert np.all(activity_coefficients[3] >= 0)


def test_fixed_water_activity():
    """test for fixed_water_activity function."""
    molar_mass_ratio = 18.016 / 250
    water_activity_desired = np.linspace(0.001, 0.9999, 1000)
    oxygen2carbon = 0.3
    density = 1.5

    _ = binary_activity.fixed_water_activity(
        water_activity=water_activity_desired,
        molar_mass_ratio=molar_mass_ratio,
        oxygen2carbon=oxygen2carbon,
        density=density,
    )



"""Operations for handling collisions between particles."""

from typing import Tuple
import torch

from particula.lagrangian import particle_pairs


def find_collisions(
        distance_matrix: torch.Tensor,
        indices: torch.Tensor,
        mass: torch.Tensor,
        k: int = 1
) -> torch.Tensor:
    """
    Find the collision pairs from a distance matrix, given the mass and
    indices of particles.

    This function identifies pairs of particles that are within a certain
    distance threshold (<0), indicating a collision.
    It filters out pairs involving particles with zero mass.

    Parameters:
    distance_matrix (torch.Tensor): A 2D tensor containing the pairwise
        distances between particles.
    indices (torch.Tensor): A 1D tensor containing the indices of the
        particles.
    mass (torch.Tensor): A 1D tensor containing the mass of each particle.
    k (int, optional): The number of closest neighbors to consider for each
        particle. Defaults to 1.

    Returns:
    torch.Tensor: A 2D tensor of shape [n_collisions, 2] containing the
    indices of colliding pairs of particles.

    Note:
    - The function assumes that the diagonal elements of the distance matrix
    (distances of particles to themselves) are less than zero.
    - Particles with zero mass are excluded from the collision pairs.
    """
    # Find the top k closest particles for each particle
    closest_neighbors = torch.topk(
        distance_matrix,
        k=k + 1,
        largest=False,
        sorted=True)

    # Identify collisions and exclude zero-distance (self) and zero-mass
    # particles
    collisions = (closest_neighbors.values < 0) * (mass.unsqueeze(1) > 0)

    # Prepare tensors for collision pairs
    iteration_index = 1  # index of the closest neighbor, only one neighbor is
    # considered for now, this could be put into a loop for multiple neighbors
    # but loops are bad, and any will just collide in the next iteration
    expanded_indices = indices.unsqueeze(1).repeat(1, k + 1)
    expanded_closest_neighbors = closest_neighbors.indices

    # Combine indices to form collision pairs
    collision_indices_pairs = torch.cat(
        [expanded_closest_neighbors.unsqueeze(2),
         expanded_indices.unsqueeze(2)], dim=2)

    # Extract valid collision pairs
    return collision_indices_pairs[
        collisions[:, iteration_index], iteration_index, :].int()


def coalescence(
        position: torch.Tensor,
        velocity: torch.Tensor,
        mass: torch.Tensor,
        radius: torch.Tensor,
        collision_indices_pairs: torch.Tensor,
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Update mass and velocity of particles based on collision pairs, conserving
    mass and momentum.

    This function processes collision pairs, sorts them to avoid duplicate
    handling, and then updates the mass and velocity of colliding particles
    according to the conservation of mass and momentum.

    Parameters:
    position (torch.Tensor): A 2D tensor of shape [n_dimensions, n_particles]
        representing the positions of particles.
    velocity (torch.Tensor): A 2D tensor of shape [n_dimensions, n_particles]
        representing the velocities of particles.
    mass (torch.Tensor): A 1D tensor containing the mass of each particle.
    radius (torch.Tensor): A 1D tensor containing the radius of each particle.
    collision_indices_pairs (torch.Tensor): A 2D tensor containing pairs of
        indices representing colliding particles.
    remove_duplicates_func (function): A function to remove duplicate entries
        from a tensor of index pairs.

    Returns:
    - torch.Tensor: A 2D tensor of shape [n_dimensions, n_particles]
        representing the updated velocities of particles.
    Note:
    - This function modifies the `velocity` and `mass` tensors in-place.
    - It assumes that the mass and momentum are transferred from the right
        particle to the left in each collision pair.
    - The subtraction approach for the right-side particles ensures no mass is
        lost in multi-particle collisions (e.g., A<-B and B<-D).
    """
    # Sort collision pairs and remove duplicate collisions (e.g. A-B and B-A)
    sorted_pairs, _ = torch.sort(collision_indices_pairs, dim=1)
    unique_left_indices = particle_pairs.remove_duplicates(sorted_pairs, 0)
    unique_indices = particle_pairs.remove_duplicates(unique_left_indices, 1)
    unique_indices = particle_pairs.validate_pair_distance(
        collision_indices_pairs=unique_indices,
        position=position,
        radius=radius,
    )
    # fast return if no collisions
    if unique_indices.shape[0] == 0:
        return velocity, mass

    # Update velocities based on conservation of momentum
    total_mass = mass[unique_indices[:, 0]] + mass[unique_indices[:, 1]]
    velocity[:, unique_indices[:, 0]] = (
        mass[unique_indices[:, 0]] * velocity[:, unique_indices[:, 0]] +
        mass[unique_indices[:, 1]] * velocity[:, unique_indices[:, 1]]
    ) / total_mass

    # Subtract mass and velocity from the right-side particles
    velocity[:, unique_indices[:, 1]] -= velocity[:, unique_indices[:, 1]]
    mass[unique_indices[:, 0]] += mass[unique_indices[:, 1]]
    mass[unique_indices[:, 1]] -= mass[unique_indices[:, 1]]

    return velocity, mass


# flake8: noqa: C901
# pylint: disable=line-too-long
def elastic_collision(
    velocity: torch.Tensor,
    mass: torch.Tensor,
    collision_indices_pairs: torch.Tensor,
) -> torch.Tensor:
    """
    Update velocities of particles based on elastic collision pairs using
    matrix operations, conserving kinetic energy and momentum.

    Parameters:
    velocity (torch.Tensor): A 2D tensor of shape [n_dimensions, n_particles]
        representing the velocities of particles.
    mass (torch.Tensor): A 1D tensor containing the mass of each particle.
    collision_indices_pairs (torch.Tensor): A 2D tensor containing pairs of
        indices representing colliding particles.
    remove_duplicates_func (function): A function to remove duplicate entries
        from a tensor of index pairs.

    Returns:
    torch.Tensor: A 2D tensor of shape [n_dimensions, n_particles]
        representing the updated velocities of particles.

    Note:
    - This function modifies the `velocity` tensor in-place.
    - Mass remains unchanged in elastic collisions.

    Example:
    2d: https://www.wolframalpha.com/input?i=elastic+collision&assumption=%7B%22F%22%2C+%22ElasticCollision%22%2C+%22m2%22%7D+-%3E%221+kg%22&assumption=%7B%22F%22%2C+%22ElasticCollision%22%2C+%22m1%22%7D+-%3E%221+kg%22&assumption=%22FSelect%22+-%3E+%7B%7B%22ElasticCollision2D%22%7D%7D&assumption=%7B%22F%22%2C+%22ElasticCollision%22%2C+%22v1i%22%7D+-%3E%221+m%2Fs%22&assumption=%7B%22F%22%2C+%22ElasticCollision%22%2C+%22v2i%22%7D+-%3E%22-0.5+m%2Fs%22
    3d fortran: https://www.plasmaphysics.org.uk/programs/coll3d_for.htm
    https://www.plasmaphysics.org.uk/collision3d.htm

    I think the approach is take a pair, and rotate the coordinate system so
    that the collision is in the x-y plane. Then, the z component of the
    velocity is a 1d problem, and the x-y component is a 2d problem. Then,
    rotate back to the original coordinate system.
    """
    # harder than in-elastic collisions

    raise NotImplementedError



"""Test the convert module."""

import numpy as np
from particula.util import stats


def test_merge_format_str_headers():
    """Test the merge_different_headers function."""
    # Create example input data
    data = np.array([[1, 2, 3], [1, 2, 3], [1, 2, 3]])
    header = ['a', 'b', 'c']
    data_add = np.array([[5, 8], [5, 8], [5, 8]])
    header_add = ['c', 'd']

    # Call function to merge the data and headers
    data, header, merged_data, merged_header = stats.merge_formatting(
        data_current=data,
        header_current=header,
        data_new=data_add,
        header_new=header_add
    )

    # Test merged data shape
    assert merged_data.shape == (3, 4)

    # Test merged header
    assert merged_header == ['a', 'b', 'c', 'd']


def test_merge_format_num_headers():
    """Test the merge_different_headers function."""
    # Create example input data
    data = np.array([[1, 2, 3], [1, 2, 3]])
    header = ['1', '2', '3']
    data_add = np.array([[5, 8], [5, 8]])
    header_add = ['3', '4']

    # Call function to merge the data and headers
    data, header, merged_data, merged_header = stats.merge_formatting(
        data_current=data,
        header_current=header,
        data_new=data_add,
        header_new=header_add
    )

    # Test merged data shape
    assert merged_data.shape == (2, 4)

    # Test merged header
    assert merged_header == ['1', '2', '3', '4']


def test_average_to_interval():
    """Test the average_to_interval function."""
    # Set up test data
    time_stream = np.arange(0, 1000, 10)
    average_base_sec = 10
    average_base_time = np.arange(0, 1000, 60)
    data_stream = np.linspace(0, 1000, len(time_stream)).reshape(-1, 1)
    data_stream = np.concatenate([data_stream, data_stream, data_stream], axis=1)
    average_base_data = np.zeros((len(average_base_time), 3))
    average_base_data_std = np.zeros((len(average_base_time), 3))

    # Call the function
    average_base_data, average_base_data_std = stats.average_to_interval(
        time_raw=time_stream,
        data_raw=data_stream,
        average_interval=average_base_sec,
        average_interval_array=average_base_time,
        average_data=average_base_data,
        average_data_std=average_base_data_std
    )
    expected_data = np.array(
            [
                [0.0, 25.252, 85.858, 146.464,
                    207.070, 267.676, 328.282, 388.888,
                    449.494, 510.101, 570.707, 631.313,
                    691.919, 752.525, 813.131, 873.737,
                    934.3434],
                [0.0, 25.252, 85.858, 146.464,
                    207.070, 267.676, 328.282, 388.888,
                    449.494, 510.101, 570.707, 631.313,
                    691.919, 752.525, 813.131, 873.737,
                    934.3434],
                [0.0, 25.252, 85.858, 146.464,
                    207.070, 267.676, 328.282, 388.888,
                    449.494, 510.101, 570.707, 631.313,
                    691.919, 752.525, 813.131, 873.737,
                    934.34343434]
            ]
        ).T

    expected_std = np.array(
            [
                [0.0, 17.250, 17.250, 17.250, 17.250,
                    17.250, 17.250, 17.250, 17.250, 17.250,
                    17.250, 17.250, 17.250, 17.250, 17.250,
                    17.250, 17.250],
                [0.0, 17.250, 17.250, 17.250, 17.250,
                    17.250, 17.250, 17.250, 17.250, 17.250,
                    17.250, 17.250, 17.250, 17.250, 17.250,
                    17.250, 17.250],
                [0.0, 17.250, 17.250, 17.250, 17.250,
                    17.250, 17.250, 17.250, 17.250, 17.250,
                    17.250, 17.250, 17.250, 17.250, 17.250,
                    17.250, 17.250]
            ]
        ).T

    assert np.allclose(average_base_data, expected_data, rtol=1e-3)
    assert np.allclose(average_base_data_std, expected_std, rtol=1e-3)


def test_mask_outliers_bottom():
    """Test the mask_outliers function."""
    data = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    bottom = 3

    expected_mask = np.array(
        [False, False, True, True, True, True, True, True, True, True])

    assert np.allclose(stats.mask_outliers(data, bottom=bottom), expected_mask)


def test_mask_outliers_top():
    """Test the mask_outliers function."""
    data = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    top = 7

    expected_mask = np.array(
        [True, True, True, True, True, True, True, False, False, False])

    assert np.allclose(stats.mask_outliers(data, top=top), expected_mask)


def test_mask_outliers_value():
    """Test the mask_outliers function."""
    data = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    value = 5

    expected_mask = np.array(
        [True, True, True, True, False, True, True, True, True, True])

    assert np.allclose(stats.mask_outliers(data, value=value), expected_mask)


def test_mask_outliers_invert():
    """Test the mask_outliers function."""
    data = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    bottom = 3
    top = 7
    invert = True

    expected_mask = np.array(
        [True, True, False, False, False, False, False, True, True, True])

    assert np.allclose(
        stats.mask_outliers(data, bottom=bottom, top=top, invert=invert),
        expected_mask)


def test_threshold_outliers_all():
    """Test the threshold_outliers function."""
    data = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    bottom = 3
    top = 7
    value = 5
    invert = True

    expected_mask = np.array(
        [True, True, False, False, True, False, False, True, True, True])

    assert np.allclose(
        stats.mask_outliers(
            data, bottom=bottom, top=top, value=value, invert=invert),
        expected_mask)



"""
Test the wall loss functions.
"""

import numpy as np

from particula.util.wall_loss import (
    spherical_wall_loss_coefficient,
    rectangle_wall_loss)


def test_spherical_wall_loss_coefficient():
    """test the spherical wall loss coefficient function"""
    # Test case with hypothetical values
    ktp_value = 1.5  # Example value
    diffusion_coefficient_value = 0.01  # Example value
    settling_velocity_value = 0.05  # Example value
    chamber_radius = 2.0  # Example value

    expected_output = 0.11816  # This is a hypothetical value.

    # Call the function with the test case
    calculated_output = spherical_wall_loss_coefficient(
        ktp_value,
        diffusion_coefficient_value,
        settling_velocity_value,
        chamber_radius
    )

    # Assertion to check if the calculated output matches the expected output
    assert np.isclose(calculated_output, expected_output, rtol=1e-4)


def test_rectangle_wall_loss():
    """test the rectangular wall loss coefficient function"""
    # Test case with hypothetical values
    ktp_value = 1.5  # Example value
    diffusion_coefficient_value = 0.01  # Example value
    settling_velocity_value = 0.05  # Example value
    dimension = [1, 1, 1]  # Example value

    expected_output = 0.47312  # This is a hypothetical value.

    # Call the function with the test case
    calculated_output = rectangle_wall_loss(
        ktp_value,
        diffusion_coefficient_value,
        settling_velocity_value,
        dimension
    )

    # Assertion to check if the calculated output matches the expected output
    assert np.isclose(calculated_output, expected_output, rtol=1e-4)



"""
Merge or adds processed data to the data stream. Accounts for data shape mis
matches and duplicate timestamps. If the data is a different shape than the
data stream, it will interpolate the data to the data stream's time array.
If the data has duplicate timestamps, it will remove the duplicates and
interpolate the data to the data stream's time array.
"""

from typing import Tuple, Optional
import numpy as np
from particula.util import convert, stats
from particula.data.stream import Stream


def combine_data(
    data: np.ndarray,
    time: np.ndarray,
    header_list: list,
    data_new: np.ndarray,
    time_new: np.ndarray,
    header_new: list,
) -> Tuple[np.ndarray, list]:
    # pylint: disable=too-many-arguments
    """"
    Merge or adds processed data together. Accounts for data shape
    miss matches and duplicate timestamps. If the data is a different shape
    than
    the existing data, it will be reshaped to match the existing data.

    Args:
    -----------
    data : np.array
        Existing data stream.
    time : np.array
        Time array for the existing data.
    header_list : List[str]
        List of headers for the existing data.
    data_new : np.array
        Processed data to add to the data stream.
    time_new : np.array
        Time array for the new data.
    header_new : List[str]
        List of headers for the new data.

    Returns:
    --------
    Tuple[np.array, List[str], Dict[str, int]]
        A tuple containing the updated data stream, the updated header list,
        and
        a dictionary mapping the header names to their corresponding indices in
        the data stream.
    """

    data_new = convert.data_shape_check(
        time=time_new,
        data=data_new,
        header=header_new)

    # Check if time_new matches the dimensions of data_new
    if np.array_equal(time, time_new):
        # no need to interpolate the data_new before adding
        # it to the data
        header_updated = np.append(header_list, header_new)
        data_updated = np.concatenate(
            (
                data,
                data_new,
            ),
            axis=1,
        )
    else:  # interpolate the data_new before adding it to the data_stream
        data_interp = np.empty((len(time), data_new.shape[1]))
        for i in range(data_new.shape[1]):
            mask = ~np.isnan(data_new[:, i])
            if not mask.any():
                data_interp[:, i] = np.nan
            else:
                left_value = data_new[mask, i][0]
                right_value = data_new[mask, i][-1]
                data_interp[:, i] = np.interp(
                    time,
                    time_new[mask],
                    data_new[mask, i],
                    left=left_value,
                    right=right_value,
                )
        # update the data array
        data_updated = np.concatenate(
            (
                data,
                data_interp,
            ),
            axis=1,
        )

    header_updated = list(np.append(header_list, header_new))
    return data_updated, header_updated


def stream_add_data(
    stream: Stream,
    time_new: np.ndarray,
    data_new: np.ndarray,
    header_check: Optional[bool] = False,
    header_new: Optional[list] = None
) -> Stream:
    """
    Adds a new data stream and corresponding time stream to the
    existing data.

    Parameters
    ----------
    stream : object
        A Stream object, containing the existing data.
    new_time : np.ndarray (m,)
        An array of time values for the new data stream.
    new_data : np.ndarray
        An array of data values for the new data stream.
    header_check : bool, optional
        If True, checks whether the header in the new data matches the
        header in the existing data. Defaults to False.
    new_header : list of str, optional
        A list of header names for the new data stream. Required if
        header_check is True.

    Returns
    -------
    stream : object
        A Stream object, containing the updated data.

    Raises
    ------
    ValueError
        If header_check is True and header is not provided or
        header does not match the existing header.

    Notes
    -----

    If header_check is True, the method checks whether the header in the
    new data matches the header in the existing data. If they do not match,
    the method attempts to merge the headers and updates the header
    dictionary.

    If header_check is False or the headers match, the new data is
    appended to the existing data.

    The function also checks whether the time stream is increasing, and if
    not, sorts the time stream and corresponding data.
    """

    if stream.data.size == 0:
        stream.data = data_new
        stream.time = time_new
    elif header_check:
        header_new = stream.header if header_new is None else header_new

        stream.data, stream.header, data_new, header_new = \
            stats.merge_formatting(
                data_current=stream.data,
                header_current=stream.header,
                data_new=data_new,
                header_new=header_new
            )
        # updates stream
        stream.data = np.vstack((stream.data, data_new))
        stream.time = np.concatenate((stream.time, time_new))
    else:
        stream.data = np.vstack((stream.data, data_new))
        stream.time = np.concatenate((stream.time, time_new))

    # check if the time stream added is increasing
    increasing_time = np.all(
        stream.time[1:] >= stream.time[:-1],
        axis=0
    )

    if not increasing_time:
        # sort the time stream
        sorted_time_index = np.argsort(stream.time)
        stream.time = stream.time[sorted_time_index]
        stream.data = stream.data[:, sorted_time_index]
    return stream



""" mobility of particle
"""
import numpy as np

from particula.util.input_handling import in_radius, in_scalar, in_viscosity
from particula.util.slip_correction import scf
from particula.util.dynamic_viscosity import dyn_vis


def pam(
    radius=None,
    scf_val=None,
    vis_val=None,
    **kwargs
):
    """ particle aerodynamic mobility
    """
    rad = in_radius(radius)
    scf_val = in_scalar(
        scf_val) if scf_val is not None else scf(radius=rad, **kwargs)

    # dyn_vis has no radius, is this a bug waiting to happen?
    vis_val = in_viscosity(
        vis_val) if vis_val is not None else dyn_vis(radius=rad, **kwargs)

    return (
        scf_val / (3 * np.pi * vis_val * rad * 2)
    )



""" testing the environment class.
"""

import pytest
from particula import environment, u

inputs = {
    "temperature": 298.15,
    "pressure": 101325
}
inputs33 = {
    "temperature": [298.15, 299, 300],
    "pressure": [101325, 101330, 101335]
}
standard_environment = environment.Environment(
    **inputs
)


def test_getters():
    """ Tests the getters for the Environment class.
    """
    assert standard_environment.temperature == inputs["temperature"] * u.K
    assert standard_environment.pressure == inputs["pressure"] * u.Pa


def test_dynamic_viscosity_air():
    """ Tests the dynamic viscosity.
    """
    assert (
        standard_environment.dynamic_viscosity().magnitude ==
        pytest.approx(1.716e-5, rel=1e-1)
    )


def test_mean_free_path_air():
    """ Test the mean free path.
    """
    assert (
        standard_environment.mean_free_path().magnitude ==
        pytest.approx(6.644e-8, rel=1e-1)
    )


def test_shapes():
    """ testing the shapes, allowing list inputs
    """

    assert (
        environment.Environment(**inputs33).pressure.m.shape
        ==
        (len(inputs33["pressure"]),)
    )
    assert (
        environment.Environment(**inputs33).dynamic_viscosity().m.shape
        ==
        (len(inputs33["pressure"]),)
    )



""" test the fuchs_sutugin calc
"""

import pytest
from particula.util.fuchs_sutugin import fsc
from particula import u


def test_fuchs_sutugin():
    """ test
    """

    assert fsc(radius=1e-9).u == u.dimensionless
    assert fsc(radius=1).m == pytest.approx(0, abs=1e-7)
    assert fsc(radius=1e-10).m == pytest.approx(1, abs=1e-3)
    assert fsc(radius=[1, 2], alpha=1).m.shape == (2,)
    assert fsc(radius=[1, 2], alpha=[1, 2]).m.shape == (2, 2)
    assert fsc(radius=1, alpha=[1, 0.9]).m.shape == (1, 2)
    assert fsc(radius=1e-6, alpha=0).m == 0



""" test the coulomb ration utility
"""

import numpy as np
import pytest
from particula import u
from particula.util.coulomb_enhancement import CoulombEnhancement as CE
from particula.util.coulomb_enhancement import cecl, cekl


def test_coulomb_ratio():
    """ testing
    """

    a_ra = u.Quantity(1, u.m)
    b_ra = u.Quantity(1, u.m)

    ce_default = CE(radius=a_ra, other_radius=b_ra)
    ce_charged = CE(radius=a_ra, other_radius=b_ra, charge=-1, other_charge=1)
    ce_extreme = CE(radius=a_ra, other_radius=b_ra,
                    charge=-1000, other_charge=1000)
    ce_unlikes = CE(radius=a_ra, radiys=b_ra, charge=-1, other_charge=-1)
    ce_neutral_a = CE(radius=a_ra, other_radius=b_ra, charge=0, other_charge=0)
    ce_neutral_b = CE(radius=a_ra, other_radius=b_ra, charge=0, other_charge=1)
    ce_neutral_c = CE(radius=a_ra, other_radius=b_ra,
                      charge=-1, other_charge=0)

    assert (
        ce_default.coulomb_potential_ratio().to_base_units().units ==
        u.dimensionless
    )
    assert (
        ce_charged.coulomb_potential_ratio().to_base_units().units ==
        u.dimensionless
    )
    assert (
        ce_neutral_a.coulomb_potential_ratio().magnitude ==
        pytest.approx(0)
    )
    assert (
        ce_charged.coulomb_potential_ratio().magnitude >=
        ce_neutral_b.coulomb_potential_ratio().magnitude
    )
    assert (
        ce_neutral_c.coulomb_potential_ratio().magnitude >=
        ce_unlikes.coulomb_potential_ratio().magnitude
    )
    assert (
        ce_extreme.coulomb_potential_ratio().magnitude ==
        pytest.approx(2e-2, rel=1e0)
    )


def test_cekl():
    """ testing cekl
    """

    a_ra = u.Quantity(1, u.m)
    b_ra = u.Quantity(1, u.m)

    ce_default = cekl(radius=a_ra, other_radius=b_ra)
    ce_charged = cekl(radius=a_ra, other_radius=b_ra,
                      charge=-1, other_charge=1)
    ce_extreme = cekl(radius=a_ra, other_radius=b_ra,
                      charge=-1000, other_charge=1000)
    ce_unlikes = cekl(radius=a_ra, radiys=b_ra, charge=-1, other_charge=-1)
    ce_neutral_a = cekl(radius=a_ra, other_radius=b_ra,
                        charge=0, other_charge=0)
    ce_neutral_b = cekl(radius=a_ra, other_radius=b_ra,
                        charge=0, other_charge=1)
    ce_neutral_c = cekl(radius=a_ra, other_radius=b_ra,
                        charge=-1, other_charge=0)

    assert (
        ce_default.to_base_units().units ==
        u.dimensionless
    )
    assert (
        ce_charged.to_base_units().units ==
        u.dimensionless
    )
    assert (
        ce_neutral_a.to_base_units().magnitude ==
        pytest.approx(1)
    )
    assert (
        ce_charged.to_base_units().magnitude >=
        ce_neutral_b.to_base_units().magnitude
    )
    assert (
        ce_neutral_c.to_base_units().magnitude >=
        ce_unlikes.to_base_units().magnitude
    )
    assert (
        ce_extreme.to_base_units().magnitude ==
        pytest.approx(1, rel=1e0)
    )


def test_cecl():
    """ testing cecm
    """

    a_ra = u.Quantity(1, u.m)
    b_ra = u.Quantity(1, u.m)

    ce_default = cecl(radius=a_ra, other_radius=b_ra)
    ce_charged = cecl(radius=a_ra, other_radius=b_ra,
                      charge=-1, other_charge=1)
    ce_extreme = cecl(radius=a_ra, other_radius=b_ra,
                      charge=-1000, other_charge=1000)
    ce_unlikes = cecl(radius=a_ra, radiys=b_ra, charge=-1, other_charge=-1)
    ce_neutral_a = cecl(radius=a_ra, other_radius=b_ra,
                        charge=0, other_charge=0)
    ce_neutral_b = cecl(radius=a_ra, other_radius=b_ra,
                        charge=0, other_charge=1)
    ce_neutral_c = cecl(radius=a_ra, other_radius=b_ra,
                        charge=-1, other_charge=0)

    assert (
        ce_default.to_base_units().units ==
        u.dimensionless
    )
    assert (
        ce_charged.to_base_units().units ==
        u.dimensionless
    )
    assert (
        ce_neutral_a.to_base_units().magnitude ==
        pytest.approx(1)
    )
    assert (
        ce_charged.to_base_units().magnitude >=
        ce_neutral_b.to_base_units().magnitude
    )
    assert (
        ce_neutral_c.to_base_units().magnitude >=
        ce_unlikes.to_base_units().magnitude
    )
    assert (
        ce_extreme.to_base_units().magnitude ==
        pytest.approx(1, rel=1e0)
    )


def test_arrays():
    """ testing if enhancements accept arrays
    """

    assert CE(
        radius=np.array([1, 2, 3]), other_radius=np.array([1, 2])
    ).coulomb_potential_ratio().m.shape == (3, 2)

    with pytest.raises(ValueError):
        CE(
            radius=[1, 2, 3],
            temperature=[1, 2, 3, 4],
        ).coulomb_potential_ratio()

    with pytest.raises(ValueError):
        CE(
            radius=[1, 2, 3],
            charge=[1, 2, 3, 4],
        ).coulomb_potential_ratio()



""" test the particle mass calculation
"""

import numpy as np
import pytest
from particula import u
from particula.util.particle_mass import mass


def test_particle_mass():
    """ Testing the particle mass calculation:

            * check defaults
            * check raise errors
            * check output units/mag
            * ensures radius is mandatory
    """

    a_radius = 1 * u.m
    a_density = 1000 * u.kg/u.m**3

    b_radius = a_radius.m

    a_mass = mass(radius=a_radius, density=a_density)
    b_mass = mass(radius=b_radius)

    assert a_mass.magnitude == 1000 * 4 * np.pi / 3
    assert a_mass.units == u.kg
    assert b_mass == a_mass

    assert mass(radius=[1e-9, 1e-8]).m.shape == (2,)
    assert mass(radius=1-9, density=[1e3, 1e4]).m.shape == (2,)
    assert mass(radius=[1e-9, 1e-8], density=[1e3, 1e4]).m.shape == (2,)

    with pytest.raises(ValueError):
        mass(radius=1*u.kg)

    with pytest.raises(ValueError):
        mass(radius=1*u.m, density=1*u.kg)

    with pytest.raises(ValueError):
        mass(radius=1*u.m, shape_factor=1*u.m)

    with pytest.raises(ValueError):
        mass(radius=1*u.m, volume_void=1*u.kg)

    with pytest.raises(ValueError):
        mass()



"""Integration methods for Lagrangian particle simulations."""

from typing import Tuple
import torch


def leapfrog(
        position: torch.Tensor,
        velocity: torch.Tensor,
        force: torch.Tensor,
        mass: torch.Tensor,
        time_step: float
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Perform a single step of leapfrog integration on the position and velocity
    of a particle.

    Leapfrog integration is a numerical method used for solving differential
    equations typically found in molecular dynamics and astrophysics.
    It is symplectic, hence conserves energy over long simulations,
    and is known for its simple implementation and stability over
    large time steps.

    Parameters:
    - position (Tensor): The current position of the particle.
    - velocity (Tensor): The current velocity of the particle.
    - force (Tensor): The current force acting on the particle.
    - mass (float): The mass of the particle.
    - time_step (float): The time step for the integration.

    Returns:
    - tuple: Updated position and velocity of the particle after one time step.

    Reference:
    - https://en.wikipedia.org/wiki/Leapfrog_integration
    """
    # Half step update for velocity
    velocity += 0.5 * time_step * force / mass
    # Full step update for position
    position += time_step * velocity
    # Another half step update for velocity
    velocity += 0.5 * time_step * force / mass
    return position, velocity



"""Test of overflow and underflow safe functions."""
import numpy as np
from particula.activity.machine_limit import safe_exp, safe_log, safe_log10


def test_safe_exp():
    """test safe_exp function."""
    # Test with positive values
    assert np.allclose(safe_exp([1, 2, 3]), np.exp([1, 2, 3]))

    # Test with negative values
    assert np.allclose(safe_exp([-1, -2, -3]), np.exp([-1, -2, -3]))

    # Test with large values
    assert not np.allclose(
        safe_exp([1000, 2000, 3000]),
        np.exp([1000, 2000, 3000]))


def test_safe_log():
    """test safe_log function."""
    # Test with positive values
    assert np.allclose(safe_log([1, 2, 3]), np.log([1, 2, 3]))

    # Test with zero values
    assert np.allclose(safe_log([0, 0, 0]), np.log(
        [np.nextafter(0, 1), np.nextafter(0, 1), np.nextafter(0, 1)]))

    # Test with negative values
    assert not np.allclose(safe_log([-1, -2, -3]), np.log([-1, -2, -3]))


def test_safe_log10():
    """test safe_log10 function."""
    # Test with positive values
    assert np.allclose(safe_log10([1, 2, 3]), np.log10([1, 2, 3]))

    # Test with zero values
    assert np.allclose(safe_log10([0, 0, 0]), np.log10(
        [np.nextafter(0, 1), np.nextafter(0, 1), np.nextafter(0, 1)]))

    # Test with negative values
    assert not np.allclose(safe_log10([-1, -2, -3]), np.log10([-1, -2, -3]))



"""Functions to operate on stream objects."""

from typing import Optional, Union
import copy
import numpy as np

from particula.util import stats
from particula.data.stream import StreamAveraged, Stream


def drop_masked(stream: Stream, mask: np.ndarray) -> Stream:
    """Drop rows where mask is false, and return data stream.

    Parameters
    ----------
    stream : object
        data stream object
    mask : np.ndarray
        mask to apply to data stream

    Returns
    -------
    object
        stream object
    """
    stream.data = stream.data[mask, :]
    stream.time = stream.time[mask]
    return stream


def average_std(
        stream: Stream,
        average_interval: Union[float, int] = 60,
        new_time_array: Optional[np.ndarray] = None,
) -> StreamAveraged:
    """
    Calculate the average and standard deviation of data within a given
    'stream' object over specified intervals.

    This function takes a 'stream' object, which should contain time-series
    data, and computes the average and standard deviation of the data at
    intervals specified by 'average_interval'. If data.time is in seconds
    then the units of the interval are seconds (hour in hours etc). The
    results are returned as a new 'StreamAveraged' object containing the
    processed data.

    Args:
    - stream (object): The input stream object containing 'time' and 'data'
        arrays along with other associated metadata.
    - average_interval (float|int, optional): The time interval over which the
        averaging is to be performed.
    - new_time_array (np.ndarray, optional): An optional array of time points
        at which the average and standard deviation are computed.
        If not provided, a new time array is generated based on the start and
        end times within the 'stream.time' object.

    Returns:
    - StreamAveraged (object): An object of type 'StreamAveraged' containing
        the averaged data, time array, start and stop times, the standard
        deviation of the averaged data, and other metadata from the original
        'stream' object.

    The function checks for an existing 'new_time_array' and generates one if
    needed. It then calculates the average and standard deviation for each
    interval and constructs a 'StreamAveraged' object with the results and
    metadata from the original 'stream' object.
    """
    # check for new time array
    if new_time_array is None:
        # generate new time array from start and end times
        new_time_array = np.arange(
            start=stream.time[0],
            stop=stream.time[-1],
            step=average_interval
        )
    # generate empty arrays for averaged data and std to be filled in
    average = np.zeros([len(new_time_array), len(stream.header)])
    std = np.zeros_like(average)

    # average data
    average, std = stats.average_to_interval(
        time_raw=stream.time,
        data_raw=stream.data,
        average_interval=average_interval,
        average_interval_array=new_time_array,
        average_data=average,
        average_data_std=std,
    )
    # write to new StreamAveraged object and return
    return StreamAveraged(
        header=stream.header,
        data=average,
        time=new_time_array,
        files=stream.files,
        average_interval=average_interval,
        start_time=new_time_array[0],
        stop_time=new_time_array[-1],
        standard_deviation=std,
    )


# pylint: disable=too-many-arguments
def filtering(
    stream: Stream,
    bottom: Optional[float] = None,
    top: Optional[float] = None,
    value: Optional[float] = None,
    invert: Optional[bool] = False,
    clone: Optional[bool] = True,
    replace_with: Optional[Union[float, int]] = None,
    drop: Optional[bool] = True,
) -> Stream:
    """
    Filters the data of the given 'stream' object based on the specified
    bounds or specific value. The filtered data can be either dropped or
    replaced with a specified value.  Note, not all parameters need to be
    specified, but at least one must be provided (top, bottom, value)

    Args:
    - stream (Stream): The input stream object containing 'data' and 'time'
        attributes.
    - bottom (float, optional): The lower bound for filtering data. Defaults
        to None.
    - top (float, optional): The upper bound for filtering data.
        Defaults to None.
    - value (float, optional): Specific value to filter from data.
        Defaults to None.
    - invert (bool): If True, inverts the filter criteria.
        Defaults to False.
    - clone (bool): If True, returns a copy of the 'stream' object, with
        filtered data. If False, modifies the 'stream' object in-place.
        Defaults to True.
    - replace_with (float|int, optional): Value to replace filtered-out data.
        Defaults to None.
    - drop (bool, optional): If True, filtered-out data points are dropped
        from the dataset. Defaults to False.

    Returns:
    - Stream: The 'stream' object with data filtered as specified.

    If 'drop' is True, 'replace_with' is ignored and filtered data points are
    removed from the 'stream' object. Otherwise, filtered data points are
    replaced with 'replace_with' value.

    add specific data row to filter on
    """
    # copy of stream object to avoid modifying original
    if clone:
        stream = copy.copy(stream)
    # Create a mask for the data that should be retained or replaced
    mask = stats.mask_outliers(
        data=stream.data,
        bottom=bottom,
        top=top,
        value=value,
        invert=invert
    )
    if drop and replace_with is None:
        # Apply mask to data and time, dropping filtered values
        # if any rows are then drop that whole column
        mask_sum = np.invert(np.sum(np.invert(mask), axis=1) > 0)
        stream = drop_masked(stream, mask_sum)
    elif replace_with is not None:
        stream.data = np.where(mask, stream.data, replace_with)
        # No need to modify 'stream.time' as it remains consistent with
        # 'stream.data'
    return stream


def remove_time_window(
        stream: Stream,
        epoch_start: Union[float, int],
        epoch_end: Optional[Union[float, int]] = None,
) -> Stream:
    """
    Remove a time window from a stream object.

    Args:
    - stream: The input stream object containing 'data' and 'time'
        attributes.
    - epoch_start: The start time of the time window to be
        removed.
    - epoch_end: The end time of the time window to be
        removed. If not provided, the time window is the closest time point to
        'epoch_start'.

    Returns:
    - Stream: The 'stream' object with the specified time window removed.
    """
    # get index of start time
    index_start = np.argmin(np.abs(stream.time - epoch_start))
    if epoch_end is None:
        # if no end time provided, remove the closest time point
        stream.time = np.delete(stream.time, index_start)
        stream.data = np.delete(stream.data, index_start, axis=0)
        return stream
    # get index of end time
    index_end = np.argmin(np.abs(stream.time - epoch_end)) + 1
    # remove time and data between start and end times
    stream.time = np.delete(stream.time, slice(index_start, index_end))
    stream.data = np.delete(stream.data, slice(index_start, index_end), axis=0)
    return stream



"""Tests for particle_pairs.py."""
import torch
from particula.lagrangian import particle_pairs


def test_remove_duplicates():
    """Test removing duplicate entries from a tensor of index pairs."""
    index_pairs = torch.tensor([[1, 2], [3, 4], [1, 2]])
    index_to_remove = 0
    expected_output = torch.tensor([[1, 2], [3, 4]])
    result = particle_pairs.remove_duplicates(index_pairs, index_to_remove)
    assert torch.all(result == expected_output)


def test_calculate_pairwise_distance():
    """Test calculating pairwise distances between points
    in a position tensor."""
    position = torch.tensor([[1, 1, 1], [1, 1, 2], [-1, 1, 1]])
    expected_output = torch.tensor([[0.0, 2.0, 2.2361],
                                    [2.0, 0.0, 1.0],
                                    [2.2361, 1.0, 0.0]])
    result = particle_pairs.calculate_pairwise_distance(position)
    assert torch.allclose(result, expected_output, atol=1e-4)


def test_single_axis_sweep_and_prune():
    """Test single axis sweep and prune algorithm."""
    # Test case with some overlapping particles
    position_axis = torch.tensor([1.0, 2.0, 4.0, 5.0])
    radius = torch.tensor([0.5, 1.5, 0.2, 0.2])
    left_indices, right_indices = particle_pairs.single_axis_sweep_and_prune(
        position_axis, radius)

    assert torch.equal(left_indices, torch.tensor([0]))
    assert torch.equal(right_indices, torch.tensor([1]))

    # Test case with no particles
    position_axis = torch.tensor([])
    radius = torch.tensor([])
    left_indices, right_indices = particle_pairs.single_axis_sweep_and_prune(
        position_axis, radius)

    assert torch.equal(left_indices, torch.tensor([]))
    assert torch.equal(right_indices, torch.tensor([]))


def test_validate_pair_distance():
    """Test validating pair distances."""
    # Mock data
    collision_indices_pairs = torch.tensor([[0, 1], [1, 2]])
    position = torch.tensor([[0.0, 1.0, 1.0],
                             [0.0, 0.0, 0.0],
                             [5.0, 5.0, 0.0]])
    radius = torch.tensor([1.5, 1.5, 0.5])

    # Expected output: Only the first pair should collide
    expected_output = torch.tensor([[0, 1]])

    # Run the function
    actual_output = particle_pairs.validate_pair_distance(
        collision_indices_pairs=collision_indices_pairs,
        position=position,
        radius=radius)

    # Assert the result
    assert torch.equal(actual_output, expected_output)



""" testing the cbar calc
"""

import pytest
from particula import u
from particula.util.rms_speed import cbar


def test_cbar():
    """ Testing the mean speed of molecules:

        1. test unitless and unit inputs
        2. test correct units
        3. test errors for invalid inputs

    """

    a_cbar = cbar()
    b_cbar = cbar(temperature=298.15)
    c_cbar = cbar(temperature=300, molecular_weight=0.03)

    assert a_cbar == b_cbar
    assert a_cbar.units == u.m/u.s
    assert c_cbar <= a_cbar

    assert cbar(temperature=[200, 300]).m.shape == (2,)
    assert cbar(molecular_weight=[0.03, 0.04]).m.shape == (2,)
    assert cbar(
        temperature=[200, 300], molecular_weight=[0.03, 0.04]
    ).m.shape == (2,)

    with pytest.raises(ValueError):
        cbar(temperature=5*u.m)

    with pytest.raises(ValueError):
        cbar(temperature=298*u.K, molecular_weight=5*u.m)



""" test friction factor calculation
"""

import pytest
from particula import u
from particula.util.friction_factor import frifac


def test_friction_factor():
    """ test the friction factor calculation

        In the continuum limit (Kn -> 0; Cc -> 1):
            6 * np.pi * dyn_vis_air * radius

        In the kinetic limit (Kn -> inf):
            8.39 * (dyn_vis_air/mfp_air) * const * radius**2

        See more: DOI: 10.1080/02786826.2012.690543 (const=1.36)
    """

    rad = 1e-9 * u.m
    dva = 1.8e-05 * u.N * u.s / u.m**2
    mfp = 65e-9 * u.m

    assert (
        frifac(radius=rad, dynamic_viscosity=dva) ==
        pytest.approx(3e-15)
    )

    assert (
        frifac(radius=rad) ==
        pytest.approx(3e-15)
    )
    assert (
        frifac(radius=rad).units ==
        (1 * u.N * u.s / u.m).to_base_units()
    )

    assert (
        frifac(radius=1e-5).magnitude ==
        pytest.approx(6*3.14*1e-5 * dva.magnitude, rel=1e-1)
    )
    assert (
        frifac(radius=1e-10).magnitude ==
        pytest.approx(8.39*1.36e-20 * (dva/mfp).magnitude)
    )

    assert frifac(radius=[1, 2, 3]).m.shape == (3,)
    assert frifac(radius=[1, 2, 3], temperature=[1, 2, 3]).m.shape == (3, 3)
    assert frifac(radius=[1, 2, 3], pressure=[1, 2, 3]).m.shape == (3, 3)



"""Test the loader module."""


import numpy as np
from particula.data import merger


def create_sample_data():
    """Create sample data for testing."""
    data = np.array([[1, 2], [1, 2], [1, 2], [1, 2], [1, 2]])
    time = np.array([0, 1, 2, 3, 4])
    header_list = ['header1', 'header2']
    return data, time, header_list


def test_combine_data_with_2d_data():
    """Test with 2d data."""
    # Setup
    data, time, header_list = create_sample_data()
    data_new = np.array([[7, 8], [7, 8], [7, 8]])
    time_new = np.array([1, 3, 4])
    header_new = ['header3', 'header4']

    # Execution
    merged_data, merged_header_list = merger.combine_data(
        data, time, header_list, data_new, time_new, header_new)

    # Verification
    expected_data = np.array([
        [1, 2, 7, 8], [1, 2, 7, 8], [1, 2, 7, 8], [1, 2, 7, 8], [1, 2, 7, 8]
    ])
    assert np.array_equal(merged_data, expected_data)
    expected_header_list = ['header1', 'header2', 'header3', 'header4']
    assert np.all(merged_header_list == expected_header_list)


def test_combine_data_with_1d_data():
    """Test with 1d data."""
    # Setup
    data, time, header_list = create_sample_data()
    data_new = np.array([7, 7])
    time_new = np.array([1, 4])
    header_new = ['header3']

    # Execution
    merged_data, merged_header_list = merger.combine_data(
        data, time, header_list, data_new, time_new, header_new)

    # Verification
    expected_data = np.array([
        [1, 2, 7], [1, 2, 7], [1, 2, 7], [1, 2, 7], [1, 2, 7]
    ])
    assert np.array_equal(merged_data, expected_data)
    expected_header_list = ['header1', 'header2', 'header3']
    assert np.all(merged_header_list == expected_header_list)


def test_combine_data_with_nan_values():
    """Test with nan values."""
    # Setup
    data, time, header_list = create_sample_data()
    data_new = np.array([
        [np.nan, 6, 7], [np.nan, 6, 7], [np.nan, 6, np.nan]])
    time_new = np.array([1, 2, 3])
    header_new = ['header3', 'header4', 'header5']

    # Execution
    merged_data, merged_header_list = merger.combine_data(
        data, time, header_list, data_new, time_new, header_new)

    # Verification
    expected_data = np.array([
        [1, 2, np.nan, 6, 7], [1, 2, np.nan, 6, 7], [1, 2, np.nan, 6, 7],
        [1, 2, np.nan, 6, 7], [1, 2, np.nan, 6, 7]
    ])
    assert np.array_equal(
        np.nan_to_num(merged_data), np.nan_to_num(expected_data)
    )
    expected_header_list = ['header1', 'header2', 'header3', 'header4',
                            'header5']
    assert np.all(merged_header_list == expected_header_list)



"""Test module for boundary.py"""

import torch
from particula.lagrangian import boundary


def test_wrapped_cube():
    """Test wrapping a position vector to a cube."""
    # Test case 1: position within the cube
    position = torch.tensor([[1.0, 2.0, 3.0], [4.0, 5.0, 4.0]])
    cube_side = 10.0
    expected_result = torch.tensor([[1.0, 2.0, 3.0], [4.0, 5.0, 4.0]])
    result = boundary.wrapped_cube(
        position=position,
        cube_side=cube_side)

    assert torch.allclose(
        result,
        expected_result)

    # Test case 2: position exceeding the positive boundary
    position = torch.tensor([[6.0, 7.0, 8.0], [5.0, 4.0, 8.0]])
    cube_side = 10.0
    expected_result = torch.tensor([[-4.0, -3.0, -2.0], [5.0, 4.0, -2.0]])
    result = boundary.wrapped_cube(
        position,
        cube_side)
    assert torch.allclose(
        result,
        expected_result)

    # Test case 3: position exceeding the negative boundary
    position = torch.tensor([[-6.0, -7.0, -8.0], [-5.0, -4.0, -8.0]])
    cube_side = 10.0
    expected_result = torch.tensor([[4.0, 3.0, 2.0], [-5.0, -4.0, 2.0]])
    assert torch.allclose(
        boundary.wrapped_cube(
            position,
            cube_side),
        expected_result)



""" calculate the vapor flux to the surface of particles
"""
import numpy as np

from particula import u
from particula.util.fuchs_sutugin import fsc as fsc_func
from particula.util.input_handling import (in_concentration, in_handling,
                                           in_scalar)
from particula.util.molecular_enhancement import mol_enh
from particula.util.particle_surface import area
from particula.util.rms_speed import cbar


def phi(  # pylint: disable=too-many-arguments
    particle_area=None,
    molecular_enhancement=None,
    vapor_attachment=1,
    vapor_speed=None,
    driving_force=1,
    fsc=None,
    **kwargs,
):
    """ vapor flux
    """
    particle_area = area(
        **kwargs) if particle_area is None else in_handling(
            particle_area, u.m**2)

    molecular_enhancement_val = mol_enh(
        **kwargs) if molecular_enhancement is None else in_scalar(
            molecular_enhancement)

    vapor_attachment = np.transpose(
        [in_scalar(in_scalar(vapor_attachment)).m]
    )*in_scalar(in_scalar(vapor_attachment)).u

    vapor_speed_val = cbar(
        **kwargs)/4 if vapor_speed is None else in_handling(
            vapor_speed, u.m/u.s)

    driving_force = in_concentration(driving_force)
    fsc_val = fsc_func(**kwargs) if fsc is None else in_scalar(fsc)

    rxd = particle_area.size
    return (
        particle_area.reshape(-1, 1) *
        molecular_enhancement_val *
        np.transpose(vapor_attachment) *
        vapor_speed_val *
        driving_force *
        fsc_val.reshape(rxd, -1)
    )



""" Testing the dimensionless coagulation module.
"""

import pytest
from particula import u
from particula.util.dimensionless_coagulation import full_coag, less_coag


def test_less():
    """ testing hardsphere dimensionless coagulation
    """

    assert (
        less_coag(radius=1e-9, other_radius=1e-8).units ==
        u.dimensionless
    )
    assert (
        less_coag(radius=1e-9, other_radius=1e-8).magnitude ==
        pytest.approx(18, rel=1e0)
    )

    assert (
        less_coag(radius=1e-9, other_radius=1e-8).magnitude <=
        less_coag(radius=1e-9, other_radius=1e-9).magnitude
    )

    assert (
        less_coag(
            radius=1e-9, other_radius=1e-8, charge=-1, other_charge=1
        ).magnitude >=
        less_coag(
            radius=1e-9, other_radius=1e-8, charge=-1, other_charge=0
        ).magnitude
    )


def test_full():
    """ testing hardsphere dimensioned coagulation
    """

    assert (
        full_coag(radius=1e-9, other_radius=1e-8).units ==
        u.m**3 / u.s
    )

    assert (
        full_coag(radius=1e-9, other_radius=1e-8).m_as(u.cm**3/u.s) ==
        pytest.approx(1.8e-8, rel=1e-1)
    )

    assert (
        full_coag(radius=1e-9, other_radius=1e-8).magnitude >=
        full_coag(radius=1e-9, other_radius=1e-9).magnitude
    )

    assert (
        full_coag(
            radius=1e-9, other_radius=1e-8, charge=-1, other_charge=1
        ).magnitude >=
        full_coag(
            radius=1e-9, other_radius=1e-8, charge=-1, other_charge=0
        ).magnitude
    )


def test_ipx():
    """ testing ion--particle coag
    """

    ion_rad = 0.5e-9
    ion_cha = -1
    ion_den = 1.84e3

    par_rad = 3e-9
    par_cha = 1
    par_den = 1.7e3

    coag = full_coag(
        radius=ion_rad, other_radius=par_rad, charge=ion_cha,
        other_charge=par_cha, density=ion_den, other_density=par_den,
        temperature=300, pressure=101325,
    )

    assert coag.m_as(u.cm**3/u.s) == pytest.approx(7.65e-8, rel=1e-1)


def test_dimensions():
    """ testing shapes, etc.
    """

    ion_rad = 0.5e-9
    ion_cha = -1
    ion_den = 1.84e3

    par_rad = 3e-9
    par_cha = 1
    par_den = 1.7e3

    par_cha_new = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    par_rad_new = [1e-9, 2e-9, 3e-9, 4e-9, 5e-9, 6e-9, 7e-9, 8e-9, 9e-9, 10e-9]
    ion_rad_new = [1e-9, 2e-9, 3e-9, 4e-9, 5e-9, 6e-9]

    coag = full_coag(
        radius=ion_rad, other_radius=par_rad, charge=ion_cha,
        other_charge=par_cha, density=ion_den, other_density=par_den,
        temperature=300, pressure=101325,
    )

    assert coag.m_as(u.cm**3/u.s).shape == (1, 1)

    with pytest.raises(ValueError):
        full_coag(
            radius=ion_rad, other_radius=par_rad, charge=ion_cha,
            other_charge=par_cha_new, density=ion_den, other_density=par_den,
            temperature=300, pressure=101325,
        )

    coag_radii = full_coag(
        radius=ion_rad, other_radius=par_rad_new, charge=ion_cha,
        other_charge=par_cha, density=ion_den, other_density=par_den,
        temperature=300, pressure=101325,
    )

    assert coag_radii.m_as(u.cm**3/u.s).shape == (1, len(par_rad_new))

    coag_ions = full_coag(
        radius=ion_rad_new, other_radius=par_rad_new, charge=ion_cha,
        other_charge=par_cha, density=ion_den, other_density=par_den,
        temperature=300, pressure=101325,
    )

    assert coag_ions.m_as(u.cm**3/u.s).shape == (
        len(ion_rad_new),
        len(par_rad_new)
    )



"""test module for particle_property.py"""
import torch
from particula.lagrangian import particle_property


def test_radius_vector():
    """Test calculating the radius vector."""
    mass = torch.tensor([10.0, 20.0, 30.0])
    density = torch.tensor([2.0, 3.0, 4.0])
    expected_radius = torch.tensor(
        [1.0608, 1.1675, 1.2143])
    result = particle_property.radius_calculation(mass, density)
    assert torch.allclose(result, expected_radius, atol=1e-4)


def test_friction_factor_runs():
    """testing if the friction factor function runs
    correct values are checked by untls tests"""
    radius = torch.tensor([1.0, 2.0, 3.0])
    temperature = 300.0
    pressure = 1e5
    result = particle_property.friction_factor_wrapper(
        radius_meter=radius,
        temperature_kelvin=temperature,
        pressure_pascal=pressure,
    )
    assert result.shape == radius.shape


def test_generate_mass():
    """test generating masses"""
    mean_radius = 100  # example value
    std_dev_radius = 15  # example value
    density = torch.tensor(1e3)  # example density
    num_particles = 1000

    masses = particle_property.generate_particle_masses(
        mean_radius=mean_radius,
        std_dev_radius=std_dev_radius,
        density=density,
        num_particles=num_particles
    )
    assert masses.shape[0] == num_particles
    assert torch.all(masses >= 0)  # Masses should be non-negative


def test_thermal_speed_calculation():
    """test calculating thermal velocity
    should check the values"""
    temperature = 273.0  # example value
    mass = particle_property.mass_calculation(radius=torch.tensor([1e-6]),
                                              density=torch.tensor([1e3]))
    expected_velocity = torch.tensor([0.0015])
    result = particle_property.thermal_speed(
        temperature_kelvin=temperature,
        mass_kg=mass,
    )
    assert torch.allclose(result, expected_velocity, atol=1e-3)


def test_speed_calculation():
    """test calculating speed"""
    velocity = torch.tensor([1.0, 2.0, 3.0])
    expected_speed = torch.tensor(3.7417)
    result = particle_property.speed(velocity)
    assert torch.allclose(result, expected_speed, atol=1e-4)


def test_random_thermal_velocity():
    """test generating thermal velocity"""
    temperature = 300  # example value
    mass = torch.tensor(1e-20)  # example mass
    num_particles = 1000

    velocities = particle_property.random_thermal_velocity(
        temperature_kelvin=temperature,
        mass_kg=mass,
        number_of_particles=num_particles
    )
    assert velocities.shape[1] == num_particles


def test_nearest_match():
    """test nearest match function"""
    # Example usage
    x = torch.tensor([0.0, 1.0, 2.0, 3.0, 4.0], dtype=torch.float)
    y = torch.tensor([0.0, 0.5, 1.0, 1.5, 2.0], dtype=torch.float)
    x_new = torch.tensor([0.5, 1.5, 2.5, 3.5], dtype=torch.float)

    # Nearest neighbor interpolation
    y_new_nearest = particle_property.nearest_match(x, y, x_new)
    expected = torch.tensor([0.0, 0.5, 1.0, 1.5])
    assert torch.allclose(y_new_nearest, expected, atol=1e-4)



"""Time management utilities."""

from datetime import datetime
import numpy as np
import pytz
from particula.util import input_handling


def time_str_to_epoch(
        time: str,
        time_format: str,
        timezone_identifier: str = 'UTC',
) -> float:
    """Convert to UTC (epoch) timezone from all inputs. Using pytz library,
    which implements the Olson time zone database. tz identifiers are strings
    from the database.
    See https://en.wikipedia.org/wiki/List_of_tz_database_time_zones
    for a list of time zones.

    Args:
    time : float (single value no arrays)
        Epoch time in seconds.
    time_format : str
        The format of the time string. See
        https://docs.python.org/3/library/datetime.html#strftime-and-strptime-format-codes
        for a list of format codes.
    timezone_identifier : str
        The time zone identifier for the current time zone.

    Returns:
    new_time : float
        The float time in the new time zone.

    Example: Date Time Format Codes
    - '2019-01-01 00:00:00' is '%Y-%m-%d %H:%M:%S'
    - '10/01/2019 00:00:00' is '%d/%m/%Y %H:%M:%S'
    - '2019-01-01 00:00:00.000000' is '%Y-%m-%d %H:%M:%S.%f'
    - '5/1/2019 1:00:00 PM' is '%m/%d/%Y %I:%M:%S %p'
    - %Y: Year with century as a decimal number.
    - %m: Month as a zero-padded decimal number.
    - %d: Day of the month as a zero-padded decimal number.
    - %H: Hour (24-hour clock) as a zero-padded decimal number.
    - %M: Minute as a zero-padded decimal number.
    - %S: Second as a zero-padded decimal number.
    - %f: Microsecond as a decimal number, zero-padded on the left.
    - %p: Locales equivalent of either AM or PM.
    # """
    time_zone = pytz.timezone(timezone_identifier)
    time_obj = datetime.strptime(time, time_format)

    time_obj = time_zone.localize(time_obj)

    return time_obj.timestamp()


def datetime64_from_epoch_array(
        epoch_array: np.ndarray,
        delta: int = 0) -> np.ndarray:
    """
    Converts an array of epoch times to a numpy array of datetime64 objects.

    Args:
    -----------
        epoch_array (np.ndarray): Array of epoch times (in seconds since
            the Unix epoch).
        delta (int): An optional offset (in seconds) to add to the epoch times
            before converting to datetime64 objects.

    Returns:
    --------
        np.ndarray: Array of datetime64 objects corresponding to the input
            epoch times.
    """
    assert len(epoch_array) > 0, "Input epoch_array must not be empty."
    # assert np.issubdtype(epoch_array.dtype, np.integer), \
    #     "Input epoch_array must be an array of integers."

    # Convert epoch times to datetime64 objects with an optional offset
    return np.array(
        [np.datetime64(int(epoch + delta), 's') for epoch in epoch_array]
    )


def relative_time(
        epoch_array: np.ndarray,
        units: str = 'hours',
) -> np.ndarray:
    """Cacluates the relative time from the start of the epoch
    array in the specified units.

    Args:
    -epoch_array (np.ndarray): Array of epoch times (in seconds since
        the Unix epoch).
    -units (str): The units of the relative time. Default is hours.

    Returns:
    -np.ndarray: Array of relative times in the specified units.
    """
    assert len(epoch_array) > 0, "Input epoch_array must not be empty."

    epoch_array -= epoch_array[0]
    epoch_array *= input_handling.convert_units('seconds', units)
    return epoch_array



""" testing the mol_enh calc
"""

import pytest
from particula import u
from particula.util.molecular_enhancement import mol_enh


def test_mol_enh():
    """ testing mol_enh
    """

    a_mol_enh = mol_enh(vapor_size=1, particle_size=1)
    b_mol_enh = mol_enh(vapor_size=1, particle_size=1)

    assert a_mol_enh == 4
    assert a_mol_enh == b_mol_enh
    assert a_mol_enh.units == u.dimensionless

    assert mol_enh(vapor_size=1, particle_size=[1, 2]).m.shape == (2, 1)
    assert mol_enh(vapor_size=[1, 2], particle_size=1).m.shape == (1, 2)
    assert mol_enh(
        vapor_size=[1, 2], particle_size=[1, 2, 3]
    ).m.shape == (3, 2)

    with pytest.raises(ValueError):
        mol_enh(vapor_size=5*u.m, particle_size=5*u.s)

    with pytest.raises(ValueError):
        mol_enh(vapor_size=1*u.kg, particle_size=5*u.m)



""" test the slip correction factor calculation
"""

import pytest
from particula import u
from particula.util.knudsen_number import knu
from particula.util.slip_correction import scf


def test_slip_correction():
    """ test the slip correction factor calculation

        the slip correction factor is approximately
            ~1      if  radius ~> 1e-6 m  (Kn -> 0)
            ~100    if  radius ~< 1e-9 m
    """

    radius_micron = 1e-6 * u.m
    radius_nano = 1e-9 * u.m
    # mean free path air
    mfp_air = 66.4e-9 * u.m
    knu_val = knu(radius=radius_micron, mfp=mfp_air)

    assert (
        scf(radius=radius_micron) ==
        pytest.approx(1, rel=1e-1)
    )

    assert (
        scf(radius=radius_nano) ==
        pytest.approx(100, rel=1e0)
    )

    assert (
        scf(radius=radius_micron, knu=knu_val) ==
        pytest.approx(1, rel=1e-1)
    )

    assert scf(radius=[1, 2, 3]).m.shape == (3,)
    assert scf(radius=1, mfp=[1, 2, 3]).m.shape == (3, 1)
    assert scf(radius=[1, 2, 3], mfp=[1, 2, 3]).m.shape == (3, 3)
    assert scf(radius=[1, 2, 3], temperature=[1, 2, 3]).m.shape == (3, 3)



""" settling velocity
"""

from particula.util.input_handling import in_radius, in_density
from particula.constants import STANDARD_GRAVITY


def psv(
    rad=None,
    den=None,
    scf_val=None,
    sgc=STANDARD_GRAVITY,
    vis_val=None,
    **kwargs
):
    """ calculate the settling velocity
    """
    _ = kwargs
    rad = in_radius(rad)
    den = in_density(den)

    return (
        (2*rad)**2 *
        den *
        scf_val *
        sgc /
        (18 * vis_val)
    )



""" testing the utility calculating reduced quantities
"""

import numpy as np
import pytest
from particula import u
from particula.util.reduced_quantity import reduced_quantity


def test_reduced_quantity():
    """ Test that the reduced quantity is calculated correctly.
    """

    assert (
        reduced_quantity(1, 2) ==
        pytest.approx(2/3)
    )
    assert (
        reduced_quantity(1 * u.kg, 2 * u.kg).magnitude ==
        pytest.approx(2/3)
    )

    assert reduced_quantity([1, 2], [2, 3]).shape == (2,)

    assert reduced_quantity(
        np.array([1, 2]), np.transpose([np.array([2, 3])])
    ).shape == (2, 2)

    with pytest.raises(TypeError):
        reduced_quantity(1, 2 * u.kg)
    with pytest.raises(TypeError):
        reduced_quantity(1 * u.kg, 2 * u.m)



"""Size distribution processing functions to calculate mean properties and
merge distributions."""

# pylint: disable=too-many-locals

from typing import Optional, List, Tuple

from math import pi
import numpy as np
from scipy.stats.mstats import gmean

from particula.util.input_handling import convert_units
from particula.util import convert
from particula.data.stream import Stream


def mean_properties(
    sizer_dndlogdp: np.ndarray,
    sizer_diameter: np.ndarray,
    total_concentration: Optional[float] = None,
    sizer_limits: Optional[list] = None
) -> Tuple[float, float, float, float, float, float, float]:
    """
    Calculates the mean properties of the size distribution.

    Parameters
    ----------
    sizer_dndlogdp : List[float]
        Concentration of particles in each bin.
    sizer_diameter : List[float]
        Bin centers
    total_concentration : Optional[float], default=None
        Total concentration of particles in the distribution.
    sizer_limits : Optional[Tuple[float, float]], default=None
        The lower and upper limits of the size of interest.

    Returns
    -------
    Tuple[float, float, float, float, float, float, float]
        Total concentration of particles in the distribution.
        Total mass of particles in the distribution.
        Mean diameter of the distribution by number.
        Mean diameter of the distribution by volume.
        Geometric mean diameter of the distribution.
        Mode diameter of the distribution by number.
        Mode diameter of the distribution by volume.
    """

    # convert to dn from dn/dlogDp
    sizer_dn = convert.convert_sizer_dn(sizer_diameter, sizer_dndlogdp)
    if total_concentration is not None:
        sizer_dn = sizer_dn * total_concentration / np.sum(sizer_dn)
    else:
        total_concentration = np.sum(sizer_dn)

    if sizer_limits is None:
        volume = 4 * pi / 3 * (sizer_diameter / 2)**3
    else:
        threshold_limits = (sizer_diameter >= sizer_limits[0]) & \
            (sizer_diameter <= sizer_limits[1])  # gets indexes to keep

        # if no bins are in the range, return nans
        if np.sum(threshold_limits) == 0:
            return np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan

        sizer_dn = sizer_dn[threshold_limits]
        sizer_diameter = sizer_diameter[threshold_limits]

        volume = 4 * pi / 3 * (sizer_diameter / 2)**3
        total_concentration = np.sum(sizer_dn)

    # Calculate mass in ug_m3 assuming a density of 1.
    mass_ug_m3 = volume * sizer_dn * 1e-9
    unit_mass_ug_m3 = np.sum(mass_ug_m3)

    # Calculate mean diameter by number.
    normalized = sizer_dn / total_concentration
    diameter_weighted = normalized * sizer_diameter
    mean_diameter_nm = np.sum(diameter_weighted)

    # mean diameter by volume, unit density is assumed so mass=volume
    normalized_vol = mass_ug_m3 / unit_mass_ug_m3
    diameter_weighted_vol = normalized_vol * sizer_diameter
    mean_vol_diameter_nm = np.sum(diameter_weighted_vol)

    # Calculate geometric mean diameter of the distribution.
    geometric_mean_diameter_nm = gmean(sizer_diameter, weights=normalized)

    # Calculate mode diameter by number.
    sizer_dn_cumsum = np.cumsum(sizer_dn) / total_concentration
    mode_diameter = np.interp(
        0.5,
        sizer_dn_cumsum,
        sizer_diameter,
        left=np.nan,
        right=np.nan
    )

    # Calculate mode diameter by mass.
    sizer_dmass_cumsum = np.cumsum(mass_ug_m3) / unit_mass_ug_m3
    mode_diameter_mass = np.interp(
        0.5,
        sizer_dmass_cumsum,
        sizer_diameter,
        left=np.nan,
        right=np.nan
    )
    return total_concentration, unit_mass_ug_m3, mean_diameter_nm, \
        mean_vol_diameter_nm, geometric_mean_diameter_nm, \
        mode_diameter, mode_diameter_mass


def sizer_mean_properties(
    stream: Stream,
    sizer_limits: Optional[List[float]] = None,
    density: float = 1.5,
    diameter_units: str = 'nm',
) -> Stream:
    """
    Calculates the mean properties of the size distribution, and returns a
    stream.

    Parameters
    ----------
    stream : Stream
        The stream to process.
    sizer_limits : list, optional [in diameter_units]
        The lower and upper limits of the size of interest. The default is None
    density : float, optional
        The density of the particles. The default is 1.5 g/cm3.
    diameter_units : str
        The units of the diameter. The default is 'nm'. This will be converted
        to nm.

    Returns
    -------
    stream : Stream
        The stream with the mean properties added.
    """

    sizer_diameter_smps = np.array(stream.header).astype(float) \
        * convert_units(diameter_units, 'nm')
    if sizer_limits is not None:  # convert to nm
        sizer_limits[0] *= convert_units(diameter_units, 'nm')
        sizer_limits[1] *= convert_units(diameter_units, 'nm')
    sizer_dndlogdp_smps = np.nan_to_num(stream.data)

    total_concentration = np.zeros_like(stream.time) * np.nan
    unit_mass_ug_m3 = np.zeros_like(total_concentration) * np.nan
    mean_diameter_nm = np.zeros_like(total_concentration) * np.nan
    mean_vol_diameter_nm = np.zeros_like(total_concentration) * np.nan
    geometric_mean_diameter_nm = np.zeros_like(total_concentration) * np.nan
    mode_diameter = np.zeros_like(total_concentration) * np.nan
    mode_diameter_mass = np.zeros_like(total_concentration) * np.nan

    total_concentration_pm01 = np.zeros_like(total_concentration) * np.nan
    unit_mass_ug_m3_pm01 = np.zeros_like(total_concentration) * np.nan

    total_concentration_pm1 = np.zeros_like(total_concentration) * np.nan
    unit_mass_ug_m3_pm1 = np.zeros_like(total_concentration) * np.nan

    total_concentration_pm25 = np.zeros_like(total_concentration) * np.nan
    unit_mass_ug_m3_pm25 = np.zeros_like(total_concentration) * np.nan

    total_concentration_pm10 = np.zeros_like(total_concentration) * np.nan
    unit_mass_ug_m3_pm10 = np.zeros_like(total_concentration) * np.nan

    for i in range(len(stream.time)):
        total_concentration[i], unit_mass_ug_m3[i], mean_diameter_nm[i], \
            mean_vol_diameter_nm[i], geometric_mean_diameter_nm[i], \
            mode_diameter[i], mode_diameter_mass[i] = \
            mean_properties(
            sizer_dndlogdp_smps[i, :],
            sizer_diameter_smps,
            sizer_limits=sizer_limits
        )

        # total PM 100 nm concentration
        total_concentration_pm01[i], unit_mass_ug_m3_pm01[i], _, _, _, _, _ = \
            mean_properties(
            sizer_dndlogdp_smps[i, :],
            sizer_diameter_smps,
            sizer_limits=[0, 100]
        )

        # total PM1 um concentration
        total_concentration_pm1[i], unit_mass_ug_m3_pm1[i], _, _, _, _, _ = \
            mean_properties(
            sizer_dndlogdp_smps[i, :],
            sizer_diameter_smps,
            sizer_limits=[0, 1000]
        )
        # total PM <2.5 um concentration
        total_concentration_pm25[i], unit_mass_ug_m3_pm25[i], _, _, _, _, _ = \
            mean_properties(
            sizer_dndlogdp_smps[i, :],
            sizer_diameter_smps,
            sizer_limits=[0, 2500]
        )

        # total PM <10 um concentration
        total_concentration_pm10[i], unit_mass_ug_m3_pm10[i], _, _, _, _, _ = \
            mean_properties(
            sizer_dndlogdp_smps[i, :],
            sizer_diameter_smps,
            sizer_limits=[0, 10000]
        )

    mass_ug_m3 = unit_mass_ug_m3 * density
    mass_ug_m3_pm01 = unit_mass_ug_m3_pm01 * density
    mass_ug_m3_pm1 = unit_mass_ug_m3_pm1 * density
    mass_ug_m3_pm25 = unit_mass_ug_m3_pm25 * density
    mass_ug_m3_pm10 = unit_mass_ug_m3_pm10 * density

    # combine the data for datalake
    combinded = np.stack((
        total_concentration,
        mean_diameter_nm,
        geometric_mean_diameter_nm,
        mode_diameter,
        mean_vol_diameter_nm,
        mode_diameter_mass,
        unit_mass_ug_m3,
        mass_ug_m3,
        total_concentration_pm01,
        unit_mass_ug_m3_pm01,
        mass_ug_m3_pm01,
        total_concentration_pm1,
        unit_mass_ug_m3_pm1,
        mass_ug_m3_pm1,
        total_concentration_pm25,
        unit_mass_ug_m3_pm25,
        mass_ug_m3_pm25,
        total_concentration_pm10,
        unit_mass_ug_m3_pm10,
        mass_ug_m3_pm10,
    ), axis=1)
    header = [
        'Total_Conc_(#/cc)',
        'Mean_Diameter_(nm)',
        'Geometric_Mean_Diameter_(nm)',
        'Mode_Diameter_(nm)',
        'Mean_Diameter_Vol_(nm)',
        'Mode_Diameter_Vol_(nm)',
        'Unit_Mass_(ug/m3)',
        'Mass_(ug/m3)',
        'Total_Conc_(#/cc)_N100',
        'Unit_Mass_(ug/m3)_N100',
        'Mass_(ug/m3)_N100',
        'Total_Conc_(#/cc)_PM1',
        'Unit_Mass_(ug/m3)_PM1',
        'Mass_(ug/m3)_PM1',
        'Total_Conc_(#/cc)_PM2.5',
        'Unit_Mass_(ug/m3)_PM2.5',
        'Mass_(ug/m3)_PM2.5',
        'Total_Conc_(#/cc)_PM10',
        'Unit_Mass_(ug/m3)_PM10',
        'Mass_(ug/m3)_PM10',
    ]

    return Stream(
        header=header,
        data=combinded,
        time=stream.time,
    )


def merge_distributions(
    concentration_lower: np.ndarray,
    diameters_lower: np.ndarray,
    concentration_upper: np.ndarray,
    diameters_upper: np.ndarray
) -> tuple[np.ndarray, np.ndarray]:
    """
    Merge two particle size distributions using linear weighting.

    Args:
    concentration_lower:
        The concentration of particles in the lower
        distribution.
    diameters_lower:
        The diameters corresponding to the lower distribution.
        concentration_upper: The concentration of particles in the upper
        distribution.
    diameters_upper:
        The diameters corresponding to the upper distribution.

    Returns:
    new_2d: The merged concentration distribution.
    new_diameter: The merged diameter distribution.

    add in an acount for the moblity vs aerodynamic diameters
    """
    # Define the linear weight function
    def weight_func(diameter, min_diameter, max_diameter):
        # Calculate the weight for each diameter
        weight = (diameter - min_diameter) / (max_diameter - min_diameter)

        # Clip the weights to the range [0, 1]
        weight = np.clip(weight, 0, 1)

        return weight

    # Find the overlapping range of diameters
    min_diameter = max(np.min(diameters_upper), np.min(diameters_lower))
    max_diameter = min(np.max(diameters_upper), np.max(diameters_lower))

    lower_min_overlap = np.argmin(np.abs(diameters_lower - min_diameter))
    upper_max_overlap = np.argmin(np.abs(diameters_upper - max_diameter))

    # Define the weighted grid
    weighted_diameter = diameters_lower[lower_min_overlap:]

    # Interpolate the lower and upper distributions onto the weighted grid
    lower_interp = concentration_lower[lower_min_overlap:]
    upper_interp = np.interp(
        weighted_diameter,
        diameters_upper,
        concentration_upper,
        left=0,
        right=0)

    # Apply the weights to the interpolated distributions
    weighted_lower = lower_interp * (
        1 - weight_func(weighted_diameter, min_diameter, max_diameter))
    weighted_upper = upper_interp * weight_func(
        weighted_diameter, min_diameter, max_diameter)

    # Add the weighted distributions together
    merged_2d = weighted_lower + weighted_upper

    # Combine the diameters
    new_diameter = np.concatenate((
        diameters_lower[:lower_min_overlap],
        weighted_diameter,
        diameters_upper[upper_max_overlap:]
    ))

    # Combine the concentrations
    new_2d = np.concatenate((
        concentration_lower[:lower_min_overlap],
        merged_2d,
        concentration_upper[upper_max_overlap:]))

    return new_2d, new_diameter


def iterate_merge_distributions(
    concentration_lower: np.ndarray,
    diameters_lower: np.ndarray,
    concentration_upper: np.ndarray,
    diameters_upper: np.ndarray
) -> tuple[np.ndarray, np.ndarray]:
    """
    Merge two sets of particle size distributions using linear weighting.

    Args:
    concentration_lower: The concentration of particles in the
        lower distribution.
    diameters_lower: The diameters corresponding to the
        lower distribution.
    concentration_upper: The concentration of particles in the
        upper distribution.
    diameters_upper: The diameters corresponding to the upper distribution.

    Returns:
    A tuple containing the merged diameter distribution and the merged
        concentration distribution.
    """
    # Iterate over all columns in the concentration datastream
    merged_2d_list = []
    for i in range(concentration_lower.shape[1]):
        # Get the current column of the lower concentration distribution
        concentration_lower_col = concentration_lower[:, i]

        # Merge the current column of the lower and upper concentration
        merged_2d, merged_diameter = merge_distributions(
            concentration_lower_col,
            diameters_lower,
            concentration_upper[:, i],
            diameters_upper
        )

        # Add the merged concentration distribution to the list
        merged_2d_list.append(merged_2d)

    # Combine the merged concentration distributions into a single array
    merged_2d_array = np.column_stack(merged_2d_list)

    # Return the merged diameter distribution and the merged concentration
    return merged_diameter, merged_2d_array


def merge_size_distribution(
    stream_lower: Stream,
    stream_upper: Stream,
    lower_units: str = 'nm',
    upper_units: str = 'um',
) -> object:
    """
    Merge two sets of particle size distributions using linear weighting.
    The concentration should be in dN/dlogDp.

    Args:
    stream_smaller: The stream with lower sizes, e.g. SMPS.
    stream_larger: The stream with larger sizes, e.g. OPS. or APS
    lower_units: The units of the lower distribution. The default is 'nm'.
    upper_units: The units of the upper distribution. The default is 'um'.

    Returns:
    Stream: A stream object containing the merged size distribution.
    """
    # Get the diameter data from the datastreams
    diameters_lower = np.array(stream_lower.header).astype(float) \
        * convert_units(lower_units, 'nm')

    diameters_upper = np.array(stream_upper.header).astype(float) \
        * convert_units(upper_units, 'nm')

    # Merge the datastreams
    merged_diameter, merged_2d = iterate_merge_distributions(
        concentration_lower=stream_lower.data,
        diameters_lower=diameters_lower,
        concentration_upper=stream_upper.data,
        diameters_upper=diameters_upper
    )
    return Stream(
        header=list(merged_diameter.astype(str)),
        data=merged_2d,
        time=stream_lower.time,
    )



""" testing vapor flux
"""

import numpy as np

from particula import u
from particula.util.vapor_flux import phi


def test_vapor_flux():
    """ testing vapor flux
    """

    assert phi(
        particle_area=1e-9*u.m**2,
        molecular_enhancement=1*u.dimensionless,
        vapor_attachment=1*u.dimensionless,
        vapor_speed=1*u.m/u.s,
        driving_force=1*u.kg/u.m**3,
        fsc=1*u.dimensionless
    ).m.size == 1

    assert phi(
        particle_area=[1e-9, 1e-9]*u.m**2,
        molecular_enhancement=np.array([[.1, .2, .3]]),
        vapor_attachment=1*u.dimensionless,
        vapor_speed=1*u.m/u.s,
        driving_force=1*u.kg/u.m**3,
        fsc=[1, 2]*u.dimensionless
    ).m.shape == (2, 3)



"""Functions to operate on stream objects."""

from typing import Optional, Union
import numpy as np

from particula.data.lake import Lake
from particula.data import stream_stats


def average_std(
        lake: Lake,
        average_interval: Union[float, int] = 60,
        new_time_array: Optional[np.ndarray] = None,
        clone: bool = True,
) -> Lake:
    """"
    Averages the data in a lake over a specified time interval.
    """
    # nice line of code
    lake_averaged = Lake() if clone else lake

    # iterate over streams in lake
    for name, stream in lake.items():
        lake_averaged[name] = stream_stats.average_std(
            stream=stream,
            average_interval=average_interval,
            new_time_array=new_time_array,
        )
    return lake_averaged



"""Test suites for particle class.
"""

import numpy as np
import pytest
from particula import environment, particle, u

small_particle = particle.Particle(
    particle_radius=1.0e-9 * u.m,
    particle_density=1.0 * u.kg / u.m**3,  # need fixing, add e3
    particle_charge=1,
)

large_particle = particle.Particle(
    name="large_particle",
    particle_radius=1.0e-7 * u.m,
    particle_density=1.8 * u.kg / u.m**3,  # need fixing, add e3
    particle_charge=1,
)

# invalid_particle = particle.Particle(
#     name="invalid_particle",
#     particle_radius=1.0e-7 * u.lb,
#     particle_density=1 * u.kg / u.m**3,
#     particle_charge=3 * u.C,
# )

standard_environment = environment.Environment(
    temperature=298 * u.K,
    pressure=101325 * u.Pa,
)

negative_ion = particle.Particle(
    name="negative_ion",
    particle_radius=0.5e-9 * u.m,
    particle_density=1.84e3 * u.kg / u.m**3,
    particle_charge=-1,
)

positive_particle = particle.Particle(
    name="positive_particle",
    particle_radius=3e-9 * u.m,
    particle_particle_density=1.7e3 * u.kg / u.m**3,
    particle_charge=1,
)

standard_environment_ip = environment.Environment(
    temperature=300 * u.K,
    pressure=101325 * u.Pa,
)

single_mode = {
    "mode": 500e-9,  # 500 nm median
    "nbins": 1000,  # 1000 bins
    "nparticles": 1e6,  # 1e4 #
    "volume": 1e-6,  # per 1e-6 m^3 (or 1 cc)
    "gsigma": 1.5,  # relatively narrow
    "spacing": "linspace",  # bin spacing,
}

multi_mode = {
    "mode": [500e-9, 500e-9],  # 500 nm median
    "nbins": 1000,  # 1000 bins
    "nparticles": [1e6, 1e6],  # 1e4 #
    "volume": 1e-6,  # per 1e-6 m^3 (or 1 cc)
    "gsigma": [1.2, 1.2],  # relatively narrow
    "spacing": "linspace",  # bin spacing,
}


def test_particle_distribution():
    """ Test that the particle distribution is calculated
        for single mode and multi mode.
    """
    def pdf_total(radius, pdf_distribution):
        return np.trapz(y=pdf_distribution, x=radius)

    # single mode
    particle_distribution_1 = particle.ParticleDistribution(
        **single_mode
    )
    total_number = pdf_total(
        particle_distribution_1.pre_radius().m,
        particle_distribution_1.pre_distribution().m,
    )
    assert total_number == pytest.approx(1e12, rel=1e10)

    # multi mode
    particle_distribution_2 = particle.ParticleDistribution(
        **multi_mode
    )
    total_number2 = pdf_total(
        particle_distribution_2.pre_radius().m,
        particle_distribution_2.pre_distribution().m
    )
    assert total_number2 == pytest.approx(2e12, rel=1e10)


def test_custom_radius():
    """ test custom radius
    """
    assert particle.Particle(radius_start=1e-9).particle_radius.m[0] == 1e-9
    assert particle.Particle(radius_end=1e-8).particle_radius.m[-1] == 1e-8
    assert particle.Particle(
        radius_start=1e-9, radius_end=1e-8).particle_radius.m[0] == 1e-9
    assert particle.Particle(
        radius_start=1e-9, radius_end=1e-8).particle_radius.m[-1] == 1e-8
    fine_spacing = particle.Particle(radius_step=1e-10)
    finer_spacing = particle.Particle(
        radius_start=1.2e-9, radius_end=15e-6, radius_step=1e-9
    )
    assert (
        fine_spacing.particle_radius.m[0] -
        fine_spacing.particle_radius.m[1]
    ) == pytest.approx(-1e-10, rel=1e-10)
    assert (
        finer_spacing.particle_radius.m[0] -
        finer_spacing.particle_radius.m[1]
    ) == pytest.approx(-1e-9, rel=1e-10)


def test_multi_mode_bins():
    """ testing the multi mode stuff for bins
    """

    assert particle.Particle(nbins=1000).particle_radius.shape == (1000,)
    assert particle.Particle(nbins=[2000]).particle_radius.shape == (2000,)
    assert particle.Particle(
        nbins=[1000, 2000]).particle_radius.shape == (3000,)


def test_getters():
    """ Test that the getters work.
    """

    assert small_particle.particle_radius == 1.0e-9 * u.m
    assert small_particle.particle_mass() == (
        4.0/3.0 * np.pi * (1.0e-9 ** 3 * u.m ** 3) * (
            1.0 * u.kg / u.m**3
        )
    )
    assert small_particle.particle_charge == 1.0


def test_individual_shapes():
    """ test individual shapes
    """
    assert particle.Particle(
        particle_radius=[1, 2]
    ).particle_radius.m.shape == (2,)
    assert particle.Particle(
        particle_radius=[1, 2]
    ).particle_mass().m.shape == (2,)
    assert particle.Particle(
        particle_radius=[1, 2]
    ).knudsen_number().m.shape == (2,)
    assert particle.Particle(
        particle_radius=[1, 2]).friction_factor().m.shape == (2,)


def test_knudsen_number():
    """ Test that the knudsen number is calculated correctly.
    """

    assert small_particle.knudsen_number() == pytest.approx(66.44, rel=1e-3)
    assert small_particle.knudsen_number().check(["None"])
    assert large_particle.knudsen_number() == pytest.approx(0.6644, rel=1e-3)
    # with pytest.raises(pint.errors.DimensionalityError):
    #     assert invalid_particle.knudsen_number() == pytest.approx(0.65)
    #     assert invalid_particle.knudsen_number().check("[None]")
    # # with pytest.raises(AssertionError):
    # #     assert invalid_particle.knudsen_number() == pytest.approx(0.65)
    # #     assert invalid_particle.knudsen_number().check("[None]")


def test_slip_correction_factor():
    """ Test that the slip correction factor is calculated correctly.
    """

    assert (
        small_particle.slip_correction_factor() ==
        pytest.approx(110.7, rel=1e-3)
    )

    assert (
        large_particle.slip_correction_factor() ==
        pytest.approx(1.886, rel=1e-3)
    )

    assert small_particle.slip_correction_factor().check(["None"])


def test_friction_factor():
    """ Test that the friction factor is calculated correctly.
    """

    assert small_particle.friction_factor(
    ).magnitude == pytest.approx(3.181e-15, rel=1e-3)

    assert (
        large_particle.friction_factor().magnitude ==
        pytest.approx(1.84e-11)
    )

    assert small_particle.friction_factor().check("[mass]/[time]")


def test_multiple_shapes():
    """ testing multiple shapes when two dists exist
    """

    assert particle.Particle(
        particle_radius=[1, 2],
    ).reduced_mass(
        other=particle.Particle(particle_radius=[1])
    ).m.shape == (2, 1)
    assert particle.Particle(
        particle_radius=[1],
    ).reduced_mass(
        other=particle.Particle(particle_radius=[1, 2])
    ).m.shape == (1, 2)

    assert particle.Particle(
        particle_radius=[1, 2],
    ).reduced_friction_factor(
        other=particle.Particle(particle_radius=[1])
    ).m.shape == (2, 1)
    assert particle.Particle(
        particle_radius=[1],
    ).reduced_friction_factor(
        other=particle.Particle(particle_radius=[1, 2])
    ).m.shape == (1, 2)

    assert particle.Particle(
        particle_radius=[1, 2],
    ).coulomb_potential_ratio(
        other=particle.Particle(particle_radius=[1])
    ).m.shape == (2, 1)
    assert particle.Particle(
        particle_radius=[1],
    ).coulomb_potential_ratio(
        other=particle.Particle(particle_radius=[1, 2])
    ).m.shape == (1, 2)

    assert particle.Particle(
        particle_radius=[1, 2],
    ).diffusive_knudsen_number(
        other=particle.Particle(particle_radius=[1])
    ).m.shape == (2, 1)
    assert particle.Particle(
        particle_radius=[1],
    ).diffusive_knudsen_number(
        other=particle.Particle(particle_radius=[1, 2])
    ).m.shape == (1, 2)

    assert particle.Particle(
        particle_radius=[1, 2],
    ).dimensionless_coagulation(
        other=particle.Particle(particle_radius=[1])
    ).m.shape == (2, 1)
    assert particle.Particle(
        particle_radius=[1],
    ).dimensionless_coagulation(
        other=particle.Particle(particle_radius=[1, 2])
    ).m.shape == (1, 2)

    assert particle.Particle(
        particle_radius=[1, 2],
    ).coagulation(
        other=particle.Particle(particle_radius=[1])
    ).m.shape == (2, 1)
    assert particle.Particle(
        particle_radius=[1],
    ).coagulation(
        other=particle.Particle(particle_radius=[1, 2])
    ).m.shape == (1, 2)


def test_big_shapes():
    """ testing big shapes
    """
    assert particle.Particle(
        particle_radius=np.linspace(1, 100, 100),
    ).dimensionless_coagulation(
        other=particle.Particle(particle_radius=[1])
    ).m.shape == (100, 1)
    assert particle.Particle(
        particle_radius=np.linspace(1, 100, 1000),
    ).dimensionless_coagulation(
        other=particle.Particle(particle_radius=np.linspace(1, 100, 500))
    ).m.shape == (1000, 500)


def test_reduced_mass():
    """ Test that the reduced mass is calculated correctly.
    """

    reduced_mass_1_2 = small_particle.reduced_mass(large_particle)
    assert reduced_mass_1_2 == pytest.approx(4.189e-27, rel=1e-3)
    assert reduced_mass_1_2.check("[mass]")


def test_reduced_friction_factor():
    """ Test that the reduced friction factor is calculated correctly.
    """

    reduced_friction_factor_1_2 = small_particle.reduced_friction_factor(
        large_particle,)
    assert reduced_friction_factor_1_2 == pytest.approx(3.18e-15)
    assert reduced_friction_factor_1_2.check("[mass]/[time]")


def test_coulomb_enh():
    """ testing coulomb enh
    """
    assert (
        small_particle.coulomb_potential_ratio(large_particle).u
        == u.dimensionless
    )
# def test_dimensionless_coagulation_kernel_parameterized():
#     """Test that the paramaterized dimensionless coagulation kernel
#     is calculated correctly.
#     """

#     assert small_particle.dimensionless_coagulation_kernel_parameterized(
#         large_particle,
#     ) == pytest.approx(0.003, rel=10)
#     assert small_particle.dimensionless_coagulation_kernel_parameterized(
#         large_particle,
#     ).check(["None"])


# def test_dimensioned_coagulation_kernel():
#     """Tests dimensioned coagulation kernel is calculated correctly.
#     """

#     assert small_particle.dimensioned_coagulation_kernel(
#         large_particle,
#     ) == pytest.approx(2.738e-10 * u.m**3 / u.s, rel=10)
#     assert small_particle.dimensioned_coagulation_kernel(
#         large_particle,
#     ).check("[length]**3/[time]")

#     # FROM PHYSICS & FIRST PRINCIPLES:
#     # when
#     # negative ion with 1 charge and size 0.5e-9 m
#     # with
#     # positive particle with 1 charge and size 3e-9 m
#     # then:
#     # coagulation should be ~1e-12 m^3/s
#     # (within 2x, or at most an order of magnitude)
#     # (conditions assumed ~300 K, ~1 atm, but don't matter much)

#     assert negative_ion.dimensioned_coagulation_kernel(
#         positive_particle,
#     ) == pytest.approx(1e-12 * u.m**3 / u.s, rel=10)
#     # rel=10 means an order of magnitude
#     assert negative_ion.dimensioned_coagulation_kernel(
#         positive_particle,
#     ).check("[length]**3/[time]")


def test_condensation_stuff():
    """ testing condensation stuff
    """

    simple_cond_kwargs = {
        "mode": 200e-9,  # 200 nm median
        "nbins": 1000,  # 1000 bins
        "nparticles": 2e6,  # 1e4 #
        "volume": 1e-6,  # per 1e-6 m^3 (or 1 cc)
        "gsigma": 1.25,  # relatively narrow
        "cutoff": .99999,  # let's take it all lol
        "vapor_radius": 1.6e-9,  # 1.6 nm
        "vapor_density": 1400,  # 1400 kg/m^3
        "vapor_concentration": 1,  # 1 ug/m^3
        "vapor_attachment": 1,  # 1
        "vapor_molecular_weight": 200,  # 200 g/mol
        "something_something": None
    }

    two_cond_vaps = simple_cond_kwargs.copy()
    two_cond_vaps["vapor_radius"] = [1.6e-9, 1.6e-9]

    two_cond_oths = simple_cond_kwargs.copy()
    two_cond_oths["vapor_radius"] = [1.6e-9, 1.7e-9]
    two_cond_oths["vapor_density"] = [1400, 1500]
    two_cond_oths["vapor_concentration"] = [1, 2]
    two_cond_oths["vapor_attachment"] = [1, 2]
    two_cond_oths["vapor_molecular_weight"] = [200, 300]

    cond = particle.Particle(**simple_cond_kwargs)
    cond2vap = particle.Particle(**two_cond_vaps)
    cond2oth = particle.Particle(**two_cond_oths)

    assert cond.particle_radius.m.shape == (1000,)
    assert cond.driving_force().u == u.kg/u.m**3
    # assert cond.driving_force().m.shape == (1,)
    assert cond.molecular_enhancement().u == u.dimensionless
    assert cond.molecular_enhancement().m.shape == (1000, 1)
    assert cond.condensation_redmass().u == u.kg/u.mol
    assert cond.condensation_redmass().m.shape == (1000, 1)
    assert cond.vapor_speed().u == u.m/u.s
    assert cond.vapor_speed().m.shape == (1000, 1)
    assert cond.vapor_flux().u == u.kg/u.s
    assert cond.vapor_flux().m.shape == (1000, 1)
    assert cond.particle_growth().u == u.m/u.s
    assert cond.particle_growth().m.shape == (1000,)

    assert cond2vap.molecular_enhancement().m.shape == (1000, 2)
    assert cond2vap.vapor_flux().m.shape == (1000, 2)
    assert cond2vap.particle_growth().m.shape == (1000,)

    assert cond2oth.molecular_enhancement().m.shape == (1000, 2)
    assert cond2oth.condensation_redmass().m.shape == (1000, 2)
    assert cond2oth.vapor_speed().m.shape == (1000, 2)

    assert cond2oth.particle_area().m.shape == (1000,)
    assert cond2oth.vapor_attachment.m.shape == (2,)
    assert cond2oth.driving_force().m.shape == (1, 2)

    assert cond2oth.fuchs_sutugin().m.shape == (1000, 2)

    assert cond2oth.vapor_flux().m.shape == (1000, 2)
    assert cond2oth.particle_growth().m.shape == (1000,)


def test_likelike_coagulation():
    """ see https://github.com/uncscode/particula/issues/239
    """
    approx_list = ["gk2008", "gh2012", "dy2007", "cg2019", "hardsphere"]
    for approx in approx_list:
        ppart = particle.Particle(
            coagulation_approximation=approx,
            particle_radius=3e-9 * u.m,
            particle_density=1.7e3 * u.kg / u.m**3,
            particle_charge=1,
        )

        assert ppart.coagulation().m >= 0



""" calculate the wall loss coefficient
"""
import numpy as np

from particula import u
from particula.util.input_handling import in_handling
from particula.util.debye_function import df1
from particula.util.settling_velocity import psv
from particula.util.diffusion_coefficient import pdc


def spherical_wall_loss_coefficient(
    ktp_value,
    diffusion_coefficient_value,
    settling_velocity_value,
    chamber_radius
):
    """Calculate the wall loss coefficient for a spherical chamber
        approximation.

    Args:
        ktp_value: rate of the wall eddy diffusivity
        diffusion_coefficient_value: Particle diffusion coefficient.
        settling_velocity_value: Settling velocity of the particle.
        chamber_radius: Radius of the chamber.

    Returns:
        The calculated wall loss coefficient for simple case.
    """
    return (
        6 * np.sqrt(ktp_value * diffusion_coefficient_value) /
        (np.pi * chamber_radius) * df1(
            np.pi * settling_velocity_value /
            (2 * np.sqrt(ktp_value * diffusion_coefficient_value))
        ) + settling_velocity_value / (4 * chamber_radius / 3)
    )


def rectangle_wall_loss(
    ktp_value,
    diffusion_coefficient_value,
    settling_velocity_value,
    dimension
):
    """
    Calculate the wall loss coefficient, β₀, for a rectangular chamber.

    Given the rate of wall diffusivity parameter (ktp_value), the particle
    diffusion coefficient (diffusion_coefficient_value), and the terminal
    settling velocity (settling_velocity_value), this function computes the
    wall loss coefficient for a rectangular-prism chamber with specified
    dimensions.

    The wall loss coefficient is calculated based on the diffusion and
    gravitational sedimentation in a rectangular chamber. It accounts for the
    effect of chamber geometry on particle loss by considering the length (L),
    width (W), and height (H) of the chamber.

    Args:
        ktp_value (float): Rate of wall diffusivity parameter in units of
            inverse seconds (s^-1).
        diffusion_coefficient_value (float): The particle diffusion
            coefficient in units of square meters per second (m^2/s).
        settling_velocity_value (float): The terminal settling velocity of the
            particles, in units of meters per second (m/s).
        dimension (tuple): A tuple of three floats representing the length (L)
            width (W), and height (H) of the rectangular chamber,
            in units of meters (m).

    Returns:
        float: The calculated wall loss coefficient (B0) for the rectangular
        chamber.

    Reference:
        The wall loss coefficient, β₀, is calculated using the following
        formula:
        $$
        \beta_0 = (LWH)^{-1} (4H(L+W) \\sqrt{k_t D}/\\pi +
        v_g LW \\coth{[(\\pi v_g)/(4\\sqrt{k_t D}})])
        $$
    """
    l, w, h = dimension  # Unpack the dimensions tuple
    # Using 1/tanh(x) for coth(x)
    coth_vg_kt_d = 1 / np.tanh(
        (np.pi * settling_velocity_value)
        / (4 * np.sqrt(ktp_value * diffusion_coefficient_value))
        )
    return (l * w * h)**-1 * (
        4 * h * (l + w) * np.sqrt(ktp_value * diffusion_coefficient_value)
        / np.pi
        + settling_velocity_value * l * w * coth_vg_kt_d)


def wlc(
    approx="none",
    ktp_value=0.1 * u.s**-1,
    diffusion_coefficient_value=None,
    dimension=1 * u.m,
    settling_velocity_value=None,
    **kwargs
):
    """Calculate the wall loss coefficient.

    Args:
        approximation: The approximation method to use, e.g., "none",
        "spherical", "rectangle"
        ktp_value: rate of the wall eddy diffusivity
        diffusion_coefficient_value: Particle diffusion coefficient.
        settling_velocity_value: Settling velocity of the particle.
        dimension: Radius of the chamber or tuple of rectangular dimensions.

    Returns:
        The calculated wall loss coefficient.
    """

    if approx == "none":
        return 0.0

    if approx in ["simple", "spherical"]:
        # input checks
        ktp_value = in_handling(ktp_value, u.s**-1)
        diffusion_coefficient_value = in_handling(
            diffusion_coefficient_value, u.m**2 / u.s
        ) if diffusion_coefficient_value is not None else pdc(**kwargs)

        settling_velocity_value = in_handling(
            settling_velocity_value, u.m / u.s
        ) if settling_velocity_value is not None else psv(**kwargs)

        chamber_radius = in_handling(
            dimension, u.m
        ) if dimension is not None else 1 * u.m

        # calculation of the wall loss coefficient
        return spherical_wall_loss_coefficient(
            ktp_value=ktp_value,
            diffusion_coefficient_value=diffusion_coefficient_value,
            settling_velocity_value=settling_velocity_value,
            chamber_radius=chamber_radius
        )

    if approx == "rectangle":
        if dimension is None:
            # Default dimensions if not provided
            dimension = (1 * u.m, 1 * u.m, 1 * u.m)
        else:
            dimension = tuple(in_handling(dim, u.m) for dim in dimension)

        return rectangle_wall_loss(
            ktp_value=ktp_value,
            diffusion_coefficient_value=diffusion_coefficient_value,
            settling_velocity_value=settling_velocity_value,
            dimension=dimension
        )

    return 0



"""Test the loader module."""


from datetime import datetime
import pytz
import numpy as np
from particula.data import loader


def test_filter_list():
    """Test the filter_list function."""
    data = ['apple,banana,orange', 'pear,kiwi,plum', 'grapefruit,lemon']
    char_counts = {',': 2}
    filtered_data = loader.filter_list(data, char_counts)
    assert filtered_data == ['apple,banana,orange', 'pear,kiwi,plum']

    data = ['apple,banana,orange', 'pear,kiwi,plum', 'grapefruit,lemon']
    char_counts = {',': 3}
    filtered_data = loader.filter_list(data, char_counts)
    assert filtered_data == []

    data = ['apple,banana,orange', 'pear,kiwi,plum', 'grapefruit,lemon']
    char_counts = {',': 1}
    filtered_data = loader.filter_list(data, char_counts)
    assert filtered_data == ['grapefruit,lemon']

    data = ['apple,banana,orange', 'pear,kiwi,plum', 'grapefruit,lemon']
    char_counts = {'z': 1}
    filtered_data = loader.filter_list(data, char_counts)
    assert filtered_data == []


def test_parse_time_column():
    """Test the parse_time_column function."""
    line = np.array('2022-01-01 12:00:00,0.5,0.6'.split(','))
    time_format = '%Y-%m-%d %H:%M:%S'

    expected_timestamp = datetime(
        2022, 1, 1, 12, 0, 0,
        tzinfo=pytz.timezone('UTC')).timestamp()

    # Test case with a single time column as integer index
    time_column_int = 0
    assert loader.parse_time_column(
        time_column=time_column_int,
        time_format=time_format,
        line=line,
    ) == expected_timestamp

    # Test case with two time columns as list of indices
    line = np.array('2022-01-01,12:00:00,0.5,0.6'.split(','))
    time_column_int = [0, 1]
    assert loader.parse_time_column(
        time_column=time_column_int,
        time_format=time_format,
        line=line,
    ) == expected_timestamp

    # Test case with invalid time column
    time_column_invalid = 3
    try:
        loader.parse_time_column(time_column_invalid, time_format, line)
        assert False, "Expected ValueError"
    except ValueError:
        assert True

    # Test case with invalid time format
    time_format_invalid = '%Y'
    try:
        loader.parse_time_column(time_column_int, time_format_invalid, line)
        assert False, "Expected ValueError"
    except ValueError:
        assert True



"""Calculates the Gibbs free energy of mixing for a binary solution."""

from particula.activity.machine_limit import safe_log


def gibbs_free_engery(
    organic_mole_fraction,
    gibbs_mix,
):
    """
    Calculate the gibbs free energy of the mixture. Ideal and non-ideal.

    Args:
    organic_mole_fraction (np.array): A numpy array of organic mole fractions.
    gibbs_mix (np.array): A numpy array of gibbs free energy of mixing.

    Returns:
    gibbs_ideal (np.array): The ideal gibbs free energy of mixing.
    gibbs_real (np.array): The real gibbs free energy of mixing.
    """

    gibbs_ideal = (1 - organic_mole_fraction) \
        * safe_log(1 - organic_mole_fraction) \
        + organic_mole_fraction \
        * safe_log(organic_mole_fraction)
    gibbs_real = gibbs_ideal + gibbs_mix
    return gibbs_ideal, gibbs_real



"""Test the time_manage module."""

import numpy as np
from particula.util import time_manage
from particula.util.time_manage import time_str_to_epoch


def test_time_str_to_epoch():
    """Test the time_str_to_epoch function."""
    # Test with a time in New York
    time = "2021-06-18 22:34:21"
    time_format = "%Y-%m-%d %H:%M:%S"
    timezone_identifier = "America/New_York"
    expected = 1624070061.0
    assert time_str_to_epoch(
        time,
        time_format,
        timezone_identifier) == expected

    # Test with a time in Los Angeles
    timezone_identifier = "America/Los_Angeles"
    expected = 1624080861.0
    assert time_str_to_epoch(
        time,
        time_format,
        timezone_identifier) == expected

    # Test with a time in London
    timezone_identifier = "Europe/London"
    expected = 1624052061.0
    assert time_str_to_epoch(
        time,
        time_format,
        timezone_identifier) == expected

    # Test with a time in Sydney
    timezone_identifier = "Australia/Sydney"
    expected = 1624019661.0
    assert time_str_to_epoch(
        time,
        time_format,
        timezone_identifier) == expected


def test_datetime64_from_epoch_array():
    """Test the datetime64_from_epoch_array function."""
    # Test with a simple example array
    epoch_array = np.array([0, 1, 2, 3, 4])
    expected_result = np.array(['1970-01-01T00:00:00', '1970-01-01T00:00:01',
                                '1970-01-01T00:00:02', '1970-01-01T00:00:03',
                                '1970-01-01T00:00:04'], dtype='datetime64[s]')
    assert np.array_equal(
            time_manage.datetime64_from_epoch_array(epoch_array),
            expected_result
        )

    # Test with a non-zero delta
    delta = 3600  # 1 hour in seconds
    expected_result = np.array(['1970-01-01T01:00:00', '1970-01-01T01:00:01',
                                '1970-01-01T01:00:02', '1970-01-01T01:00:03',
                                '1970-01-01T01:00:04'], dtype='datetime64[s]')
    assert np.array_equal(
            time_manage.datetime64_from_epoch_array(epoch_array, delta),
            expected_result
        )

    # Test with an empty array
    empty_array = np.array([])
    try:
        time_manage.datetime64_from_epoch_array(empty_array)
        assert False, \
            "Function should raise an AssertionError for empty array."
    except AssertionError:
        pass

    # Test with a non-integer array
    float_array = np.array([0.0, 1.0, 2.0])
    try:
        time_manage.datetime64_from_epoch_array(float_array)
        assert False, \
            "Function should raise an AssertionError for non-integer array."
    except AssertionError:
        pass



""" defining the environment class

    kwargs:
        coagulation_approx...   (string)         (default: "hardsphere")
        dilution_rate_constant  (float) [1/s]    (default: 0.)
        temperature             (float) [K]      (default: 298.15)
        pressure                (float) [Pa]     (default: 101325)
        dynamic_viscosity       (float) [Pa*s]   (default: util)
        molecular_weight        (float) [kg/mol] (default: constants)
        reference_viscosity     (float) [Pa*s]   (default: constants)
        reference_temperature   (float) [K]      (default: constants)
        sutherland_constant     (float) [K]      (default: constants)
        gas_constant            (float) [J/mol/K](default: constants)

      Using particula.util:
        dynamic_viscosity       (float) [Pa*s]

      Using particula.constants:
        GAS_CONSTANT            (float) [J/mol/K]
        MOLECULAR_WEIGHT_AIR    (float) [kg/mol]
        REF_VISCOSITY_AIR_STP   (float) [Pa*s]
        REF_TEMPERATURE_STP     (float) [K]
        SUTHERLAND_CONSTANT     (float) [K]

"""

from particula import u
from particula.constants import (GAS_CONSTANT, MOLECULAR_WEIGHT_AIR,
                                 REF_TEMPERATURE_STP, REF_VISCOSITY_AIR_STP,
                                 SUTHERLAND_CONSTANT)
from particula.util.dynamic_viscosity import dyn_vis
from particula.util.input_handling import (in_gas_constant, in_handling,
                                           in_molecular_weight, in_pressure,
                                           in_temperature, in_viscosity,
                                           in_scalar)
from particula.util.mean_free_path import mfp
from particula.util.dilution_loss import drc
from particula.util.species_properties import (vapor_concentration)


class SharedProperties:  # pylint: disable=too-few-public-methods
    """ a hidden class for sharing properties like
        coagulation_approximation
    """

    def __init__(self, **kwargs):
        """ initiate
        """

        self.coagulation_approximation = str(
            kwargs.get("coagulation_approximation", "hardsphere")
        )
        self.dilution_rate_constant = in_handling(
            kwargs.get("dilution_rate_coefficient", 0.0),
            u.s**-1
        )

    def dilution_rate_coefficient(self):
        """ get the dilution rate coefficient
        """

        return drc(
            value=self.dilution_rate_constant
        )


class Environment(
    SharedProperties
):  # pylint: disable=too-many-instance-attributes
    """ creating the environment class

        For now, the environment class takes properties such as
        temperature and pressure to calculate derived properties
        such as viscosity and mean free path.
    """

    def __init__(self, **kwargs):
        """ Initiate the environment class with base attrs.
        """

        super().__init__(**kwargs)

        self.temperature = in_temperature(
            kwargs.get("temperature", 298.15)
        )
        self.reference_viscosity = in_viscosity(
            kwargs.get("reference_viscosity", REF_VISCOSITY_AIR_STP)
        )
        self.reference_temperature = in_temperature(
            kwargs.get("reference_temperature", REF_TEMPERATURE_STP)
        )
        self.pressure = in_pressure(
            kwargs.get("pressure", 101325)
        )
        self.molecular_weight = in_molecular_weight(
            kwargs.get("molecular_weight", MOLECULAR_WEIGHT_AIR)
        )
        self.sutherland_constant = in_temperature(
            kwargs.get("sutherland_constant", SUTHERLAND_CONSTANT)
        )
        self.gas_constant = in_gas_constant(
            kwargs.get("gas_constant", GAS_CONSTANT)
        )
        self.water_saturation_ratio = in_scalar(
            kwargs.get("water_activity", 0.0)
        )  # saturation ratio = water activity = relative humidity / 100

        self.kwargs = kwargs

    def dynamic_viscosity(self):
        """ Returns the dynamic viscosity in Pa*s.
        """
        return dyn_vis(
            temperature=self.temperature,
            reference_viscosity=self.reference_viscosity,
            reference_temperature=self.reference_temperature,
            sutherland_constant=self.sutherland_constant,
        )

    def mean_free_path(self):
        """ Returns the mean free path in m.
        """
        return mfp(
            temperature=self.temperature,
            pressure=self.pressure,
            molecular_weight=self.molecular_weight,
            dynamic_viscosity=self.dynamic_viscosity(),
            gas_constant=self.gas_constant,
        )

    def water_vapor_concentration(self):
        """ Returns the water vapor concentration in kg/m^3.
        """
        return vapor_concentration(
                saturation_ratio=self.water_saturation_ratio,
                temperature=self.temperature,
                species="water"
            )



"""Particle self calculation. e.g. radius, surface area, volume, etc."""

import numpy as np
import torch

from particula.constants import BOLTZMANN_CONSTANT
from particula.util import friction_factor, dynamic_viscosity, \
    slip_correction, mean_free_path, knudsen_number
from particula.util.input_handling import convert_units


def radius_calculation(
        mass: torch.Tensor,
        density: torch.Tensor
) -> torch.Tensor:
    """
    Calculate the radius of a sphere given its mass and density using the
    formula for the volume of a sphere.

    This function assumes a uniform density and spherical shape to compute the
    radius based on the mass-density relationship:
    Volume = Mass / Density. The volume of a sphere is given by
    (4/3) * pi * radius^3.

    Args:
    - mass (torch.Tensor): A tensor containing the mass of the sphere(s). Can
        be a scalar or a vector.
    - density (torch.Tensor): A tensor containing the density of the sphere(s).
        Can be a scalar or a vector.

    Returns:
    torch.Tensor: A tensor of the same shape as `mass` and `density`
        representing the radius of the sphere(s).

    Note:
    - The function supports broadcasting, so `mass` and `density` can be of
        different shapes, as long as they are broadcastable to a common shape.
    - Units of mass and density should be consistent to obtain a radius in
        meaningful units.
    """
    return torch.pow(3 * mass / (4 * np.pi * density), 1 / 3)


def mass_calculation(
        radius: torch.Tensor,
        density: torch.Tensor
) -> torch.Tensor:
    """
    Calculate the mass of a sphere given its radius and density using the
    formula for the volume of a sphere.

    This function assumes a uniform density and spherical shape to compute the
    mass based on the mass-density relationship:
    Volume = Mass / Density. The volume of a sphere is given by
    (4/3) * pi * radius^3.

    Args:
    - radius (torch.Tensor): A tensor containing the radius of the sphere(s).
        Can be a scalar or a vector.
    - density (torch.Tensor): A tensor containing the density of the sphere(s).
        Can be a scalar or a vector.

    Returns:
    torch.Tensor: A tensor of the same shape as `radius` and `density`
        representing the mass of the sphere(s).
    """
    return 4 * np.pi * radius ** 3 * density / 3


def friction_factor_wrapper(
        radius_meter: torch.Tensor,
        temperature_kelvin: float,
        pressure_pascal: float,
) -> torch.Tensor:
    """
    Calculate the friction factor for a given radius, temperature, and
    pressure.

    This function wraps several underlying calculations related to
    dynamic viscosity, mean free path, Knudsen number, and slip correction
    factor to compute the particle friction factor.

    Args:
        radius_meter: A tensor representing the radius of the
        sphere(s) in meters. Can be a scalar or a vector.
        temperature_kelvin: A tensor of the temperature in Kelvin.
        pressure_pascal: A tensor of the pressure in Pascals.

    Returns:
        torch.Tensor: A tensor of the same shape as `radius_meter`,
        representing the particle friction factor.
    """
    # get dynamic viscosity
    dynamic_viscosity_value = dynamic_viscosity.dyn_vis(
        temperature_kelvin=temperature_kelvin,
        pressure_pascal=pressure_pascal,
    )
    # get mean free path
    mean_free_path_meter = mean_free_path.mfp(
        temperature_kelvin=temperature_kelvin,
        pressure_pascal=pressure_pascal,
        dynamic_viscosity=dynamic_viscosity_value,
    )
    # get knudsen number
    knudsen = knudsen_number.knu(
        radius=radius_meter,
        mfp=mean_free_path_meter,
    )
    # get slip correction factor
    slip_correction_factor = slip_correction.scf(
        radius=radius_meter,
        knu=knudsen,
    )

    return friction_factor.frifac(
        radius=radius_meter,
        dynamic_viscosity=dynamic_viscosity_value,
        scf_val=slip_correction_factor,
    ).m


def generate_particle_masses(
        mean_radius: float,
        std_dev_radius: float,
        density: torch.Tensor,
        num_particles: int,
        radius_input_units: str = "nm",
) -> torch.Tensor:
    """
    Generate an array of particle masses based on a log-normal distribution of
    particle radii and a given density.

    Args:
        mean_radius (float): Mean radius of the particles. The units are
            specified by `radius_input_units`.
        std_dev_radius (float): Standard deviation of the particle radii. The
            units are specified by `radius_input_units`.
        density (torch.Tensor): Density of the particles in kg/m^3.
        num_particles (int): Number of particles to generate.
        radius_input_units (str, optional): Units of `mean_radius` and
            `std_dev_radius`. Defaults to 'nm' (nanometers).

    Returns:
        torch.Tensor: A tensor of particle masses in kg, corresponding to each
            particle.

    Raises:
        ValueError: If `mean_radius` or `std_dev_radius` are non-positive.
    """
    if mean_radius <= 0 or std_dev_radius <= 0:
        raise ValueError(
            "Mean radius and standard deviation must be positive.")

    # Convert mean and standard deviation from the specified units to meters
    mean_log = np.log(mean_radius)
    std_dev_log = np.log(std_dev_radius)
    if mean_log < 0 or std_dev_log < 0:
        # need to check on this error, not clear why this happens
        raise ValueError(
            "log of Mean radius and standard deviation must be positive for"
            + " torch.distributions.log_normal.LogNormal")

    # Sample radii from the log-normal distribution
    radius_samples = torch.distributions.log_normal.LogNormal(
        mean_log, std_dev_log).sample((num_particles,))
    radius_samples *= convert_units(radius_input_units, "m")

    # Calculate mass of each particle
    return mass_calculation(radius=radius_samples, density=density)


def thermal_speed(
        temperature_kelvin: float,
        mass_kg: torch.Tensor,
) -> torch.Tensor:
    """
    Calculate the thermal speed of a particle based on its temperature and
    mass.

    The thermal speed is computed using the formula: sqrt(8 * k * T / (pi * m))
    where k is the Boltzmann constant, T is the temperature in Kelvin, and m is
    the particle mass in kilograms.

    Args:
        temperature_kelvin (float): Temperature of the environment in Kelvin.
        mass_kg (torch.Tensor): Mass of the particle(s) in kilograms.
            Can be a scalar or a vector.

    Returns:
        torch.Tensor: The thermal speed of the particle(s) in meters per second

    Raises:
        ValueError: If `temperature_kelvin` is less than or equal to zero or
        if any element of `mass_kg` is non-positive.
    """
    if temperature_kelvin <= 0:
        raise ValueError("Temperature must be greater than zero.")
    if torch.any(mass_kg <= 0):
        raise ValueError("All mass values must be positive.")

    return torch.sqrt(8 * BOLTZMANN_CONSTANT.magnitude *
                      temperature_kelvin / (np.pi * mass_kg))


def speed(
    velocity: torch.Tensor,
) -> torch.Tensor:
    """
    Calculate the speed of a particle.

    Args:
        velocity (torch.Tensor): Velocity of the particle.

    Returns:
        torch.Tensor: Speed of the particle.
    """
    return torch.sqrt(torch.sum(velocity**2, dim=0))


def random_thermal_velocity(
    temperature_kelvin: float,
    mass_kg: torch.Tensor,
    number_of_particles: int,
    t_type=torch.float,
    random_seed: int = 0,
) -> torch.Tensor:
    """
    Generate a random thermal velocity for each particle.

    Args:
        temperature_kelvin (torch.Tensor): Temperature of the fluid in Kelvin.
        mass_kg (torch.Tensor): Mass of the particle in kilograms.
        number_of_particles (int): Number of particles.

    Returns:
        torch.Tensor: Thermal speed of the particle in meters per second.
    """
    # set the seed
    torch.manual_seed(random_seed)
    # Initialize particle velocities uniformly random
    unit_velocity = torch.rand(3, number_of_particles, dtype=t_type) - 0.5
    # get the speed
    unit_speed = speed(unit_velocity)
    # get the thermal speed
    thermal_particle_speed = thermal_speed(
        temperature_kelvin=temperature_kelvin,
        mass_kg=mass_kg,
    )
    # normalize the unit
    return unit_velocity * thermal_particle_speed / unit_speed


def nearest_match(
    x_values: torch.Tensor,
    y_values: torch.Tensor,
    x_new: torch.Tensor,
) -> torch.Tensor:
    """
    Perform nearest neighbor interpolation (on torch objects) to find y-values
    corresponding to new x-values. The function identifies the nearest x-value
    for each value in x_new and returns the corresponding y-value.

    Args:
        x_values (torch.Tensor): The original x-values of shape (n,).
        y_values (torch.Tensor): The original y-values of shape (n,).
            Each y-value corresponds to an x-value.
        x_new (torch.Tensor): The new x-values for which y-values are to be
            interpolated, of shape (m,).

    Returns:
        torch.Tensor: The interpolated y-values of shape (m,). Each value
            corresponds to the nearest match from x_values.
    """
    # Reshape x_values and x_new for broadcasting
    x_values = x_values.unsqueeze(0)
    x_new = x_new.unsqueeze(1)
    # Find the indices of the nearest x-values in x_values for each value in
    # x_new
    idx = torch.argmin(torch.abs(x_values - x_new), dim=1)
    # Return the corresponding y-values for the found indices
    return y_values[idx]



"""Lagrangian particle pairwise distances and pairwise operations."""

from typing import Tuple
import torch


def remove_duplicates(
        index_pairs: torch.Tensor,
        index_to_remove: int
) -> torch.Tensor:
    """
    Removes duplicate entries from a specified column in a tensor of index
    pairs.

    This function is designed to work with tensors where each row represents a
    pair of indices. It removes rows containing duplicate entries in the
    specified column.

    Parameters:
    - index_pairs (torch.Tensor): A 2D tensor of shape [n, 2], where n is the
        number of index pairs.
    - index_to_remove (int): The column index (0 or 1) from which to remove
        duplicate entries.

    Returns:
    - torch.Tensor: A 2D tensor of index pairs with duplicates removed from
        the specified column.

    Example:
    index_pairs = torch.tensor([[1, 2], [3, 4], [1, 2]])
    index_to_remove = 0
    # Output will be [[1, 2], [3, 4]] assuming column 0 is chosen for removing
        duplicates.
    """
    # Sort index_pairs by the index_to_remove column
    sorted_index_pairs, indices_sorted = torch.sort(
        index_pairs[:, index_to_remove], dim=0)

    # Find unique entries in the index_to_remove column
    # diff_index is True for unique entries
    diff_index = torch.diff(sorted_index_pairs, prepend=torch.tensor([-1])) > 0

    # Select rows from index_pairs that correspond to unique entries
    clean_index = indices_sorted[diff_index]
    return index_pairs[clean_index]


def calculate_pairwise_distance(position: torch.Tensor) -> torch.Tensor:
    """
    need to test this:

    Calculate the pairwise Euclidean distances between points in a given
    position tensor.

    This function computes the pairwise distances between points represented
    in the input tensor. Each row of the input tensor is considered a point in
    n-dimensional space.

    Parameters:
    position (torch.Tensor): A 2D tensor of shape [n_dimensions, n_points]

    Returns:
    torch.Tensor: A 2D tensor of shape [n_points, n_points] containing the
    pairwise Euclidean distances between each pair of points.
    The element at [i, j] in the output tensor represents the distance
    between the i-th and j-th points in the input tensor.

    Example:
    position = torch.tensor([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    # Output will be a 3x3 tensor with the pairwise distances between these
    3 points.
    """
    # Expand position tensor to compute pairwise differences
    detla_position = position.unsqueeze(2) - position.unsqueeze(1)
    # Compute pairwise Euclidean distances
    return torch.sqrt(torch.sum(detla_position**2, dim=0))


def validate_pair_distance(
    collision_indices_pairs: torch.Tensor,
    position: torch.Tensor,
    radius: torch.Tensor
) -> torch.Tensor:
    """
    Validates if the Euclidean distances between pairs of points are smaller
    than the sum of their radii.

    Args:
        collision_indices_pairs (torch.Tensor): A tensor containing pairs of
            indices of potentially colliding particles.
        position (torch.Tensor): A 2D tensor of particle positions, where each
            column represents a particle, and each row represents an axis.
        radius (torch.Tensor): A 1D tensor representing the radius of each
            particle.

    Returns:
        torch.Tensor: A tensor containing the indices of the pairs of
            particles that are actually colliding.
    """
    # Fast return if there are no particle pairs
    if collision_indices_pairs.numel() == 0:
        return torch.tensor([], dtype=torch.bool)

    # Calculate 3D distance for each pair of particles
    delta_position = position[:, collision_indices_pairs[:, 0]] \
        - position[:, collision_indices_pairs[:, 1]]
    # Euclidean distance between particles
    distance = torch.sqrt(torch.sum(delta_position**2, axis=0))
    # radius sum of both particles
    distance_threshold = radius[collision_indices_pairs[:, 0]] \
        + radius[collision_indices_pairs[:, 1]]

    # Return the pairs of particles where the distance is less than the sum of
    # their radii
    return collision_indices_pairs[distance < distance_threshold]


def single_axis_sweep_and_prune(
    position_axis: torch.Tensor,
    radius: torch.Tensor
) -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Sweep and prune algorithm for collision detection along a single axis.
    This function identifies pairs of particles that are close enough to
    potentially collide.

    Args:
        position_axis (torch.Tensor): The position of particles along a single
            axis.
        radius (torch.Tensor): The radius of particles.

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: Two tensors containing the indices
        of potentially colliding particles.
    """

    # Fast return if there are no particles
    if position_axis.shape[0] == 0:
        return torch.tensor([], dtype=torch.int64), torch.tensor(
            [], dtype=torch.int64)

    # Apply sweep and prune to find pairs of particles that are close enough
    # to collide
    sweep = torch.sort(position_axis)
    sweep_diff = torch.diff(sweep.values)
    radius_sum = radius[sweep.indices[:-1]] + radius[sweep.indices[1:]]

    # Select indices of particles that are close enough to collide
    prune_bool = sweep_diff < radius_sum
    left_overlap_indices = sweep.indices[torch.cat(
        [prune_bool, torch.tensor([False], dtype=torch.bool)])]
    right_overlap_indices = sweep.indices[torch.cat(
        [torch.tensor([False], dtype=torch.bool), prune_bool])]

    return left_overlap_indices, right_overlap_indices


# pylint: disable=too-many-locals
def full_sweep_and_prune(
        position: torch.Tensor,
        radius: torch.Tensor
) -> torch.Tensor:
    """
    Sweep and prune algorithm for collision detection along all three axes
    (x, y, z). This function identifies pairs of particles that are close
    enough to potentially collide in 3D space.

    Args:
        position (torch.Tensor): The 2D tensor of particle positions,
            where each row represents an axis (x, y, z).
        radius (torch.Tensor): The radius of particles.

    Returns:
        torch.Tensor: A tensor containing pairs of indices of potentially
            colliding particles.
    """
    # select only particles with positive radius
    valid_radius = radius > 0
    valid_radius_indices = torch.arange(radius.shape[0])[valid_radius]
    # sweep x axis
    left_x_overlap_shifted, right_x_overlap_shifted = \
        single_axis_sweep_and_prune(
            position_axis=position[0, valid_radius_indices],
            radius=radius[valid_radius_indices]
        )
    # fast return if there are no particles overlapping in x
    if left_x_overlap_shifted.shape[0] == 0:
        return torch.tensor([])
    # unshift from relative valid radius to position index
    left_x_overlap_indices = valid_radius_indices[left_x_overlap_shifted]
    right_x_overlap_indices = valid_radius_indices[right_x_overlap_shifted]

    # cobine left and right indices for next step
    all_overlaps_x = torch.cat(
        [left_x_overlap_indices, right_x_overlap_indices])
    # select unique indices
    indices_x_unique = torch.unique(all_overlaps_x)

    # sweep y axis
    left_y_overlap_shifted, right_y_overlap_shifted = \
        single_axis_sweep_and_prune(
            position_axis=position[1][indices_x_unique],
            radius=radius[indices_x_unique]
        )
    # fast return if there are no particles overlapping in y
    if left_y_overlap_shifted.shape[0] == 0:
        return torch.tensor([])
    # unshift from x relative index to position index
    left_y_overlap_indices = indices_x_unique[left_y_overlap_shifted]
    right_y_overlap_indices = indices_x_unique[right_y_overlap_shifted]

    # combine left and right indices for next step
    all_overlaps_y = torch.cat(
        [left_y_overlap_indices, right_y_overlap_indices])
    # select unique indices
    indices_y_unique = torch.unique(all_overlaps_y)

    # sweep z axis
    left_z_overlap_shifted, right_z_overlap_shifted = \
        single_axis_sweep_and_prune(
            position_axis=position[2][indices_y_unique],
            radius=radius[indices_y_unique]
        )
    # fast return if there are no particles overlapping in z
    if left_z_overlap_shifted.shape[0] == 0:
        return torch.tensor([])
    # unshift from y relative index to position index
    left_z_overlap_indices = indices_y_unique[left_z_overlap_shifted]
    right_z_overlap_indices = indices_y_unique[right_z_overlap_shifted]

    return torch.cat(
        [
            left_z_overlap_indices.unsqueeze(1),
            right_z_overlap_indices.unsqueeze(1),
        ],
        dim=1,
    )


def full_sweep_and_prune_simplified(
    position: torch.Tensor,
    radius: torch.Tensor,
    working_yet: bool = False
) -> torch.Tensor:
    """
    A simplified version of the full sweep and prune algorithm for collision
    written above, it is not working yet. there is an error in the update of
    the indices in the y and z axis.

    Sweep and prune algorithm for collision detection along all three axes
    (x, y, z). This function identifies pairs of particles that are close
    enough to potentially collide in 3D space.

    Args:
        position (torch.Tensor): The 2D tensor of particle positions,
            where each row represents an axis (x, y, z).
        radius (torch.Tensor): The radius of particles.

    Returns:
        torch.Tensor: A tensor containing pairs of indices of potentially
            colliding particles.
    """
    if not working_yet:
        raise NotImplementedError
    if radius.shape[0] == 0:
        return torch.tensor([])

    valid_indices = torch.arange(radius.shape[0])[radius > 0]
    unique_indices = valid_indices

    for axis in range(position.shape[0]):
        if unique_indices.shape[0] == 0:
            break

        left_overlap_indices, right_overlap_indices = \
            single_axis_sweep_and_prune(
                position_axis=position[axis, unique_indices],
                radius=radius[unique_indices]
            )

        if left_overlap_indices.shape[0] == 0:
            return torch.tensor([])

        # Combine indices to form collision pairs, may still have duplicates
        all_overlaps = torch.cat(
            [unique_indices[left_overlap_indices],
             unique_indices[right_overlap_indices]])
        if axis < position.shape[0]:  # not last axis
            unique_indices = torch.unique(all_overlaps)  # remove duplicates

    return torch.cat(
        [
            unique_indices[left_overlap_indices].unsqueeze(1),
            unique_indices[right_overlap_indices].unsqueeze(1),
        ],
        dim=1,
    )



"""A module for the Stream and StreamAveraged(Stream) classes."""


from typing import List, Union
from dataclasses import dataclass, field
import numpy as np
from particula.util import time_manage


@dataclass
class Stream:
    """A class for consistent data storage and format.

    Attributes:
    ---------
    header : List[str]
        A list of strings representing the header of the data stream.
    data : np.ndarray
        A numpy array representing the data stream. The first dimension
        represents time and the second dimension represents the header.
    time : np.ndarray
        A numpy array representing the time stream.
    files : List[str]
        A list of strings representing the files containing the data stream.

    Methods:
    -------
    validate_inputs
        Validates the inputs to the Stream class.
    datetime64 -> np.ndarray
        Returns an array of datetime64 objects representing the time stream.
        Useful for plotting, with matplotlib.dates.
    return_header_dict -> dict
        Returns the header as a dictionary with keys as header elements and
        values as their indices.
    """

    # Initialize other fields as empty arrays
    header: List[str] = field(default_factory=list)
    data: np.ndarray = field(default_factory=lambda: np.array([]))
    time: np.ndarray = field(default_factory=lambda: np.array([]))
    files: List[str] = field(default_factory=list)

    def __post_init__(self):
        self.validate_inputs()

    def validate_inputs(self):
        """
        Validates the inputs for the DataStream object.
        Raises:
            TypeError: If header is not a list.
        # this might be why I can't call Stream without inputs
        """
        if not isinstance(self.header, list):
            raise TypeError("header_list must be a list")

    def __getitem__(self, index: Union[int, str]):
        """Allows for indexing of the data stream.
        Args:
        ----------
        index : int or str
            The index of the data stream to return.
        Returns:
        -------
        np.ndarray
            The data stream at the specified index."""
        if isinstance(index, str):
            index = self.header.index(index)
        return self.data[:, index]

    def __setitem__(self, index: Union[int, str], value):
        """Allows for setting or adding of a row of data in the stream.
        Args:
        index : The index of the data stream to set.
        value : The data to set at the specified index.

        future work maybe add a list option and iterate through the list"""
        if isinstance(index, str):
            if index not in self.header:
                self.header.append(index)  # add new header element
                self.data = np.hstack((self.data, value))
            index = self.header.index(index)
        # if index is an int, set the data at that index
        self.data[:, index] = value

    def __len__(self):
        """Returns the length of the time stream."""
        return len(self.time)

    @property
    def datetime64(self) -> np.ndarray:
        """
        Returns an array of datetime64 objects representing the time stream.
        Useful for plotting, with matplotlib.dates.
        """
        return time_manage.datetime64_from_epoch_array(self.time)

    @property
    def header_dict(self) -> dict:
        """Returns the header as a dictionary with index (0, 1) as the keys
        and the names as values."""
        return dict(enumerate(self.header))

    @property
    def header_float(self) -> np.ndarray:
        """Returns the header as a numpy array of floats."""
        return np.array(self.header, dtype=float)


@dataclass
class StreamAveraged(Stream):
    """A subclass of Stream with additional parameters related to averaging.

    Attributes:
        average_interval (float): The size of the window used for averaging.
        start_time (float): The start time for averaging.
        stop_time (float): The stop time for averaging.
        standard_deviation (float): The standard deviation of the data.
    """

    average_interval: float = field(default_factory=float)
    start_time: float = field(default_factory=float)
    stop_time: float = field(default_factory=float)
    standard_deviation: np.ndarray = field(
        default_factory=lambda: np.array([]))

    def __post_init__(self):
        super().__post_init__()
        self.validate_averaging_params()

    def validate_averaging_params(self):
        """
        Validates the averaging parameters for the stream.

        Raises:
            ValueError: If average_window is not a positive number or if
            start_time and stop_time are not numbers or if start_time is
            greater than or equal to stop_time.
        """
        if not isinstance(self.average_interval, (int, float)
                          ) or self.average_interval <= 0:
            raise ValueError("average_window must be a positive number")
        if not isinstance(self.start_time, (int, float)) or \
                not isinstance(self.stop_time, (int, float)) or \
                self.start_time >= self.stop_time:
            raise ValueError(
                "start_time must be less than stop_time and numerical")

    def get_std(self, index) -> np.ndarray:
        """Returns the standard deviation of the data."""
        if isinstance(index, str):
            index = self.header.index(index)
        return self.standard_deviation[:, index]



""" testing the dynamic viscosity function
"""

import pytest
from particula import u
from particula.constants import REF_TEMPERATURE_STP as REF_TEMP
from particula.constants import REF_VISCOSITY_AIR_STP as REF_VIS_AIR
from particula.util.dynamic_viscosity import dyn_vis


def test_dyn_vis():
    """ Testing the dynamic viscosity calculation:
            * see if dyn_vis returns correct units of constant
            * see if defaults work properly
            * see if scaling is correct (dividing by 2)
            * see if conversion from celsius works
            * see if user-provided constants work properly
            * testing invalid inputs

    """

    assert dyn_vis().units == REF_VIS_AIR.to_base_units().units

    assert dyn_vis(temperature=REF_TEMP) == REF_VIS_AIR

    assert dyn_vis(temperature=REF_TEMP) == REF_VIS_AIR

    assert dyn_vis(temperature=REF_TEMP/2) <= REF_VIS_AIR

    assert dyn_vis(temperature=REF_TEMP*2) >= REF_VIS_AIR

    assert (
        dyn_vis(temperature=REF_TEMP.to("degC")).to_base_units()
        ==
        REF_VIS_AIR.to_base_units()
    )

    assert (
        dyn_vis(
            temperature=REF_TEMP,
            reference_viscosity=REF_VIS_AIR/2,
            reference_temperature=REF_TEMP,
        )
        ==
        REF_VIS_AIR/2
    )

    assert dyn_vis(temperature=[250, 255, 260]).m.shape == (3,)

    with pytest.raises(ValueError):
        dyn_vis(temperature=5*u.m)

    with pytest.raises(ValueError):
        dyn_vis(reference_viscosity=5*u.m)

    with pytest.raises(ValueError):
        dyn_vis(reference_temperature=5*u.m)



""" calculate kelvin correction """

import numpy as np

from particula import u
from particula.util.input_handling import in_temperature, in_density, in_radius
from particula.util.input_handling import in_handling, in_molecular_weight
from particula.constants import GAS_CONSTANT


def kelvin_radius(
    surface_tension=0.072 * u.N/u.m,
    molecular_weight=0.01815 * u.kg/u.mol,
    density=1000 * u.kg/u.m**3,
    temperature=298.15 * u.K
):
    """ Kelvin radius (Neil's definition)
        https://en.wikipedia.org/wiki/Kelvin_equation
    """

    temperature = in_temperature(temperature).to_base_units()
    molecular_weight = in_molecular_weight(molecular_weight).to_base_units()
    density = in_density(density).to_base_units()
    surface_tension = in_handling(surface_tension, u.N/u.m)  # type: ignore

    return 2 * surface_tension * molecular_weight / (
        GAS_CONSTANT * temperature * density
    )


def kelvin_term(radius=None, **kwargs):
    """ Kelvin term (Neil's definition)
        https://en.wikipedia.org/wiki/Kelvin_equation
    """

    rad = in_radius(radius)

    kelvin_rad = kelvin_radius(**kwargs)

    return np.exp(kelvin_rad / rad)



""" testing for the vapor class
"""

from particula import u
from particula.vapor import Vapor

inputs = {
    "vapor_radius": 1.5e-9,
    "vapor_density": 1500,
    "temperature": 300,
    "pressure": 1e5,
}

inputs2 = {
    "vapor_radius": [1.5e-9, 1.6e-9],
    "vapor_density": [1500, 1600],
    "vapor_molecular_weight": [200, 300],
    "vapor_concentration": [1, 2],
}

vaps = Vapor(**inputs)
vaps2 = Vapor(**inputs2)


def test_basics():
    """ testing basic stuff
    """
    assert vaps.vapor_radius == inputs["vapor_radius"] * u.m
    assert vaps.vapor_density == inputs["vapor_density"] * u.kg/u.m**3
    assert vaps.vapor_molec_wt.u == u.kg/u.mol


def test_shapes():
    """ testing sizes and shapes
    """
    assert vaps2.vapor_density.m.shape == (2,)
    assert vaps2.vapor_molec_wt.m.shape == (2,)
    assert vaps2.driving_force().m.shape == (1, 2)


def test_units():
    """ testing units
    """
    assert vaps2.vapor_density.u == u.kg/u.m**3
    assert vaps2.vapor_molec_wt.u == u.kg/u.mol
    assert vaps2.driving_force().u == u.kg/u.m**3


def test_inheritance():
    """ testing inheritance from Environment class
    """
    assert vaps.temperature.u == u.K
    assert vaps.pressure.u == u.kg/u.m/u.s**2
    assert vaps.temperature.m == inputs["temperature"]
    assert vaps.pressure.m == inputs["pressure"]
    assert vaps.dynamic_viscosity().u == (1 * u.Pa * u.s).to_base_units()
    assert vaps.mean_free_path().u == u.m
